###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################
import inspect
from functools import reduce
from typing import Any, ClassVar, Union
from collections.abc import Callable

import txaio

from autobahn import wamp
from autobahn.exception import PayloadExceededError
from autobahn.util import IdGenerator, ObservableMixin, public
from autobahn.wamp import exception, message, role, types, uri
from autobahn.wamp.exception import (
    ApplicationError,
    ProtocolError,
    SerializationError,
    TypeCheckError,
)
from autobahn.wamp.interfaces import (
    IAuthenticator,
    IMessage,
    IPayloadCodec,
    ISession,
    ITransport,
)  # noqa
from autobahn.wamp.request import (
    CallRequest,
    Endpoint,
    Handler,
    InvocationRequest,
    Publication,
    PublishRequest,
    RegisterRequest,
    Registration,
    SubscribeRequest,
    Subscription,
    UnregisterRequest,
    UnsubscribeRequest,
)
from autobahn.wamp.types import (
    CallResult,
    Challenge,
    CloseDetails,
    EncodedPayload,
    RegisterOptions,
    SessionDetails,
    SubscribeOptions,
)


def is_method_or_function(f):
    return inspect.ismethod(f) or inspect.isfunction(f)


class BaseSession(ObservableMixin):
    """
    WAMP session base class.

    This class implements :class:`autobahn.wamp.interfaces.ISession`.
    """

    log = None

    def __init__(self):
        self.log = txaio.make_logger()

        self.set_valid_events(
            valid_events=[
                "join",  # right before onJoin runs
                "leave",  # after onLeave has run
                "ready",  # after onJoin and all 'join' listeners have completed
                "connect",  # right before onConnect
                "disconnect",  # right after onDisconnect
            ]
        )

        # this is for marshalling traceback from exceptions thrown in user
        # code within WAMP error messages (that is, when invoking remoted
        # procedures)
        self.traceback_app = False

        # mapping of exception classes to list of WAMP error URI patterns
        self._ecls_to_uri_pat: dict[ClassVar, list[uri.Pattern]] = {}

        # mapping of WAMP error URIs to exception classes
        self._uri_to_ecls: dict[str, ClassVar] = {
            ApplicationError.INVALID_PAYLOAD: SerializationError,
            ApplicationError.PAYLOAD_SIZE_EXCEEDED: PayloadExceededError,
        }

        # WAMP ITransport (_not_ a Twisted protocol, which is self.transport - when using Twisted)
        self._transport: ITransport | None = None

        # session authentication information
        self._realm: str | None = None
        self._session_id: int | None = None
        self._authid: str | None = None
        self._authrole: str | None = None
        self._authmethod: str | None = None
        self._authprovider: str | None = None
        self._authextra: dict[str, Any] | None = None

        # set client role features supported and announced
        self._session_roles: dict[str, role.RoleFeatures] = role.DEFAULT_CLIENT_ROLES

        # complete session details
        self._session_details: SessionDetails | None = None

        # payload transparency codec
        self._payload_codec: IPayloadCodec | None = None

        # generator for WAMP request IDs
        self._request_id_gen = IdGenerator()

    @property
    def transport(self) -> ITransport | None:
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.transport`
        """
        # Note: self._transport (which is a WAMP ITransport) is different from self.transport (which
        # is a Twisted protocol when using Twisted)!
        return self._transport

    @public
    def is_connected(self) -> bool:
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.is_connected`
        """
        return self._transport is not None and self._transport.isOpen()

    @public
    def is_attached(self) -> bool:
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.is_attached`
        """
        return self._session_id is not None and self.is_connected()

    @property
    def session_details(self) -> SessionDetails | None:
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.session_details`
        """
        return self._session_details

    @property
    def realm(self) -> str | None:
        return self._realm

    @property
    def session_id(self) -> int | None:
        return self._session_id

    @property
    def authid(self) -> str | None:
        return self._authid

    @property
    def authrole(self) -> str | None:
        return self._authrole

    @property
    def authmethod(self) -> str | None:
        return self._authmethod

    @property
    def authprovider(self) -> str | None:
        return self._authprovider

    @property
    def authextra(self) -> dict[str, Any] | None:
        return self._authextra

    def define(self, exception: Exception, error: str | None = None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.define`
        """
        if error is None:
            if hasattr(exception, "_wampuris"):
                self._ecls_to_uri_pat[exception] = exception._wampuris
                self._uri_to_ecls[exception._wampuris[0].uri()] = exception
            else:
                raise RuntimeError(
                    'cannot define WAMP exception from class with no decoration ("_wampuris" unset)'
                )
        else:
            if not hasattr(exception, "_wampuris"):
                self._ecls_to_uri_pat[exception] = [
                    uri.Pattern(error, uri.Pattern.URI_TARGET_HANDLER)
                ]
                self._uri_to_ecls[error] = exception
            else:
                raise RuntimeError(
                    'cannot define WAMP exception: error URI is explicit, but class is decorated ("_wampuris" set'
                )

    def _message_from_exception(
        self, request_type, request, exc, tb=None, enc_algo=None
    ):
        """
        Create a WAMP error message from an exception.

        :param request_type: The request type this WAMP error message is for.
        :type request_type: int

        :param request: The request ID this WAMP error message is for.
        :type request: int

        :param exc: The exception.
        :type exc: Instance of :class:`Exception` or subclass thereof.

        :param tb: Optional traceback. If present, it'll be included with the WAMP error message.
        :type tb: list or None
        """
        args = None
        if hasattr(exc, "args"):
            args = list(exc.args)  # make sure tuples are made into lists

        kwargs = None
        if hasattr(exc, "kwargs"):
            kwargs = exc.kwargs

        if tb:
            if kwargs:
                kwargs["traceback"] = tb
            else:
                kwargs = {"traceback": tb}

        if isinstance(exc, exception.ApplicationError):
            error = exc.error if type(exc.error) == str else exc.error
        else:
            if exc.__class__ in self._ecls_to_uri_pat:
                error = self._ecls_to_uri_pat[exc.__class__][0]._uri
            else:
                error = "wamp.error.runtime_error"

        encoded_payload = None
        if self._payload_codec:
            encoded_payload = self._payload_codec.encode(False, error, args, kwargs)

        if encoded_payload:
            msg = message.Error(
                request_type,
                request,
                error,
                payload=encoded_payload.payload,
                enc_algo=encoded_payload.enc_algo,
                enc_key=encoded_payload.enc_key,
                enc_serializer=encoded_payload.enc_serializer,
            )
        else:
            msg = message.Error(request_type, request, error, args, kwargs)

        return msg

    def _exception_from_message(self, msg):
        """
        Create a user (or generic) exception from a WAMP error message.

        :param msg: A WAMP error message.
        :type msg: instance of :class:`autobahn.wamp.message.Error`
        """

        # FIXME:
        # 1. map to ecls based on error URI wildcard/prefix
        # 2. extract additional args/kwargs from error URI

        exc = None
        enc_err = None

        if msg.enc_algo:
            if not self._payload_codec:
                log_msg = "received encoded payload, but no payload codec active"
                self.log.warn(log_msg)
                enc_err = ApplicationError(
                    ApplicationError.ENC_NO_PAYLOAD_CODEC,
                    log_msg,
                    enc_algo=msg.enc_algo,
                )
            else:
                try:
                    encoded_payload = EncodedPayload(
                        msg.payload, msg.enc_algo, msg.enc_serializer, msg.enc_key
                    )
                    decrypted_error, msg.args, msg.kwargs = self._payload_codec.decode(
                        True, msg.error, encoded_payload
                    )
                except Exception as e:
                    self.log.warn(
                        "failed to decrypt application payload 1: {err}", err=e
                    )
                    enc_err = ApplicationError(
                        ApplicationError.ENC_DECRYPT_ERROR,
                        f"failed to decrypt application payload 1: {e}",
                        enc_algo=msg.enc_algo,
                    )
                else:
                    if msg.error != decrypted_error:
                        self.log.warn(
                            "URI within encrypted payload ('{decrypted_error}') does not match the envelope ('{error}')",
                            decrypted_error=decrypted_error,
                            error=msg.error,
                        )
                        enc_err = ApplicationError(
                            ApplicationError.ENC_TRUSTED_URI_MISMATCH,
                            f"URI within encrypted payload ('{decrypted_error}') does not match the envelope ('{msg.error}')",
                            enc_algo=msg.enc_algo,
                        )

        if enc_err:
            return enc_err

        if msg.error in self._uri_to_ecls:
            ecls = self._uri_to_ecls[msg.error]
            try:
                # the following might fail, eg. TypeError when
                # signature of exception constructor is incompatible
                # with args/kwargs or when the exception constructor raises
                if msg.kwargs:
                    if msg.args:
                        exc = ecls(*msg.args, **msg.kwargs)
                    else:
                        exc = ecls(**msg.kwargs)
                else:
                    if msg.args:
                        exc = ecls(*msg.args)
                    else:
                        exc = ecls()
            except Exception:
                try:
                    self.onUserError(
                        txaio.create_failure(),
                        "While re-constructing exception",
                    )
                except:
                    pass

        if not exc:
            # the following ctor never fails ..
            if msg.kwargs:
                if msg.args:
                    exc = exception.ApplicationError(msg.error, *msg.args, **msg.kwargs)
                else:
                    exc = exception.ApplicationError(msg.error, **msg.kwargs)
            else:
                if msg.args:
                    exc = exception.ApplicationError(msg.error, *msg.args)
                else:
                    exc = exception.ApplicationError(msg.error)

        # FIXME: cleanup and integate into ctors above
        if hasattr(exc, "enc_algo"):
            exc.enc_algo = msg.enc_algo
        if hasattr(exc, "callee"):
            exc.callee = msg.callee
        if hasattr(exc, "callee_authid"):
            exc.callee_authid = msg.callee_authid
        if hasattr(exc, "callee_authrole"):
            exc.callee_authrole = msg.callee_authrole
        if hasattr(exc, "forward_for"):
            exc.forward_for = msg.forward_for

        return exc


@public
class ApplicationSession(BaseSession):
    """
    WAMP application session for applications (networking framework agnostic parts).

    Implements (partially):

        * :class:`autobahn.wamp.interfaces.ITransportHandler`
        * :class:`autobahn.wamp.interfaces.ISession`
    """

    def __init__(self, config: types.ComponentConfig | None = None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession`
        """
        BaseSession.__init__(self)
        self._config: types.ComponentConfig = config or types.ComponentConfig(
            realm="realm1"
        )

        # for keeping transport-closing state
        self._goodbye_sent: bool = False
        self._transport_is_closing: bool = False

        # WAMP application payload codec (e.g. for WAMP-cryptobox E2E)
        self._payload_codec: IPayloadCodec | None = None

        # outstanding requests
        self._publish_reqs = {}
        self._subscribe_reqs = {}
        self._unsubscribe_reqs = {}
        self._call_reqs = {}
        self._register_reqs = {}
        self._unregister_reqs = {}

        # subscriptions in place
        self._subscriptions = {}

        # registrations in place
        self._registrations = {}

        # incoming invocations
        self._invocations = {}

    @property
    def config(self) -> types.ComponentConfig:
        return self._config

    @public
    def set_payload_codec(self, payload_codec: IPayloadCodec | None):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.set_payload_codec`
        """
        self._payload_codec = payload_codec

    @public
    def get_payload_codec(self) -> IPayloadCodec | None:
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.get_payload_codec`
        """
        return self._payload_codec

    @public
    def onOpen(self, transport: ITransport):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onOpen`
        """
        self.log.debug(
            "{func}(transport={transport})", func=self.onOpen, transport=transport
        )

        # The WAMP transport (e.g. WebSocket connection)
        self._transport = transport

        # FIXME: the observer API gets "transport" as argument, but _not_ the onConnect callback below?
        d = self.fire("connect", self, transport)
        txaio.add_callbacks(
            d, None, lambda fail: self._swallow_error(fail, "While notifying 'connect'")
        )
        txaio.add_callbacks(
            d,
            lambda _: txaio.as_future(self.onConnect),
            lambda fail: self._swallow_error(fail, "While calling 'onConnect'"),
        )

    @public
    def onConnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onConnect`
        """
        self.log.debug("{func}()", func=self.onConnect)
        self.join(self.config.realm)

    @public
    def join(
        self,
        realm: str,
        authmethods: list[str] | None = None,
        authid: str | None = None,
        authrole: str | None = None,
        authextra: dict[str, Any] | None = None,
        resumable: bool | None = None,
        resume_session: int | None = None,
        resume_token: str | None = None,
    ):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.join`
        """
        if self._session_id:
            raise Exception("session already joined")

        if not self._transport:
            raise Exception("no transport set for session")

        # store the realm requested by client, though this might be overwritten later,
        # when realm redirection kicks in
        self._realm = realm

        # closing handshake state
        self._goodbye_sent = False

        # send HELLO message to router
        msg = message.Hello(
            realm=realm,
            roles=self._session_roles,
            authmethods=authmethods,
            authid=authid,
            authrole=authrole,
            authextra=authextra,
            resumable=resumable,
            resume_session=resume_session,
            resume_token=resume_token,
        )
        self._transport.send(msg)

    @public
    def disconnect(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.disconnect`
        """
        if self._transport:
            self._transport.close()

    @public
    def onUserError(self, fail, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
        """
        if hasattr(fail, "value") and isinstance(
            fail.value, exception.ApplicationError
        ):
            self.log.warn(
                '{klass}.onUserError(): "{msg}"',
                klass=self.__class__.__name__,
                msg=fail.value.error_message(),
            )
        else:
            self.log.error(
                '{klass}.onUserError(): "{msg}"\n{traceback}',
                klass=self.__class__.__name__,
                msg=msg,
                traceback=txaio.failure_format_traceback(fail),
            )

    def _swallow_error(self, fail, msg):
        """
        This is an internal generic error-handler for errors encountered
        when calling down to on*() handlers that can reasonably be
        expected to be overridden in user code.

        Note that it *cancels* the error, so use with care!

        Specifically, this should *never* be added to the errback
        chain for a Deferred/coroutine that will make it out to user
        code.
        """
        try:
            self.onUserError(fail, msg)
        except Exception:
            self.log.error(
                "Internal error: {tb}",
                tb=txaio.failure_format_traceback(txaio.create_failure()),
            )
        return None

    def type_check(self, func):
        """
        Does parameter type checking and validation against type hints
        and appropriately tells the user code and the caller (through router).
        """

        async def _type_check(*args, **kwargs):
            # Converge both args and kwargs into a dictionary
            arguments = inspect.getcallargs(func, *args, **kwargs)
            response = []
            for name, kind in func.__annotations__.items():
                if name in arguments:
                    # if a "type" has __origin__ attribute it is a
                    # parameterized generic
                    if getattr(kind, "__origin__", None) == Union:
                        expected_types = [arg.__name__ for arg in kind.__args__]
                        if not isinstance(arguments[name], kind.__args__):
                            response.append(
                                f"'{name}' expected types={expected_types} got={type(arguments[name]).__name__}"
                            )
                    elif not isinstance(arguments[name], kind):
                        response.append(
                            f"'{name}' expected type={kind.__name__} got={type(arguments[name]).__name__}"
                        )
            if response:
                raise TypeCheckError(", ".join(response))
            return await txaio.as_future(func, *args, **kwargs)

        return _type_check

    def onMessage(self, msg: IMessage):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onMessage`
        """

        if self._session_id is None:
            # the first message must be WELCOME, ABORT or CHALLENGE ..
            if isinstance(msg, message.Welcome):
                # before we let user code see the session -- that is,
                # before we fire "join" -- we give authentication
                # instances a chance to abort the session. Usually
                # this would be for "mutual authentication"
                # scenarios. For example, WAMP-SCRAM uses this to
                # confirm the server-signature
                d = txaio.as_future(self.onWelcome, msg)

                def success(res):
                    if res is not None:
                        self.log.debug("Session denied by onWelcome: {res}", res=res)
                        reply = message.Abort(
                            "wamp.error.cannot_authenticate", f"{res}"
                        )
                        self._transport.send(reply)
                        return

                    if msg.realm:
                        self._realm = msg.realm
                    self._session_id = msg.session
                    self._authid = msg.authid
                    self._authrole = msg.authrole
                    self._authmethod = msg.authmethod
                    self._authprovider = msg.authprovider
                    self._authextra = msg.authextra
                    self._router_roles = msg.roles

                    self._session_details = SessionDetails(
                        realm=self._realm,
                        session=self._session_id,
                        authid=self._authid,
                        authrole=self._authrole,
                        authmethod=self._authmethod,
                        authprovider=self._authprovider,
                        authextra=msg.authextra,
                        serializer=self._transport._serializer.SERIALIZER_ID,
                        transport=self._transport.transport_details,
                        # FIXME
                        resumed=False,  # msg.resumed,
                        resumable=False,  # msg.resumable,
                        resume_token=None,  # msg.resume_token,
                    )

                    # firing 'join' *before* running onJoin, so that
                    # the idiom where you "do stuff" in onJoin --
                    # possibly including self.leave() -- works
                    # properly. Besides, there's "ready" that fires
                    # after 'join' and onJoin have all completed...
                    d = self.fire("join", self, self._session_details)
                    # add a logging errback first, which will ignore any
                    # errors from fire()
                    txaio.add_callbacks(
                        d,
                        None,
                        lambda e: self._swallow_error(e, "While notifying 'join'"),
                    )
                    # this should run regardless
                    txaio.add_callbacks(
                        d,
                        lambda _: txaio.as_future(self.onJoin, self._session_details),
                        None,
                    )
                    # ignore any errors from onJoin (XXX or, should that be fatal?)
                    txaio.add_callbacks(
                        d, None, lambda e: self._swallow_error(e, "While firing onJoin")
                    )
                    # this instance is now "ready"...
                    txaio.add_callbacks(d, lambda _: self.fire("ready", self), None)
                    # ignore any errors from 'ready'
                    txaio.add_callbacks(
                        d,
                        None,
                        lambda e: self._swallow_error(e, "While notifying 'ready'"),
                    )

                def error(e):
                    reply = message.Abort(
                        "wamp.error.cannot_authenticate",
                        "Error calling onWelcome handler",
                    )
                    self._transport.send(reply)
                    return self._swallow_error(e, "While firing onWelcome")

                txaio.add_callbacks(d, success, error)

            elif isinstance(msg, message.Abort):
                # fire callback and close the transport
                details = types.CloseDetails(msg.reason, msg.message)
                d = txaio.as_future(self.onLeave, details)

                def success(arg):
                    # XXX also: handle async
                    d = self.fire("leave", self, details)

                    def return_arg(_):
                        return arg

                    def _error(e):
                        return self._swallow_error(e, "While firing 'leave' event")

                    txaio.add_callbacks(d, return_arg, _error)
                    return d

                def _error(e):
                    return self._swallow_error(e, "While firing onLeave")

                txaio.add_callbacks(d, success, _error)

            elif isinstance(msg, message.Challenge):
                challenge = types.Challenge(msg.method, msg.extra)
                d = txaio.as_future(self.onChallenge, challenge)

                def success(signature):
                    if signature is None:
                        raise Exception(
                            "onChallenge user callback did not return a signature"
                        )
                    if type(signature) == bytes:
                        signature = signature.decode("utf8")
                    if type(signature) != str:
                        raise Exception(
                            f"signature must be unicode (was {type(signature)})"
                        )
                    reply = message.Authenticate(signature)
                    self._transport.send(reply)

                def error(err):
                    self.onUserError(err, "Authentication failed")
                    reply = message.Abort(
                        "wamp.error.cannot_authenticate", f"{err.value}"
                    )
                    self._transport.send(reply)
                    # fire callback and close the transport
                    details = types.CloseDetails(reply.reason, reply.message)
                    d = txaio.as_future(self.onLeave, details)

                    def success(arg):
                        # XXX also: handle async
                        self.fire("leave", self, details)
                        return arg

                    def _error(e):
                        return self._swallow_error(e, "While firing onLeave")

                    txaio.add_callbacks(d, success, _error)
                    # switching to the callback chain, effectively
                    # cancelling error (which we've now handled)
                    return d

                txaio.add_callbacks(d, success, error)

            else:
                raise ProtocolError(
                    f"Received {msg.__class__} message, and session is not yet established"
                )

        else:
            # self._session_id != None (aka "session established")
            if isinstance(msg, message.Goodbye):
                if not self._goodbye_sent:
                    # the peer wants to close: send GOODBYE reply
                    reply = message.Goodbye()
                    self._transport.send(reply)

                self._session_id = None

                # fire callback and close the transport
                details = types.CloseDetails(msg.reason, msg.message)
                d = txaio.as_future(self.onLeave, details)

                def success(arg):
                    # XXX also: handle async
                    self.fire("leave", self, details)
                    return arg

                def _error(e):
                    errmsg = f'While firing onLeave for reason "{msg.reason}" and message "{msg.message}"'
                    return self._swallow_error(e, errmsg)

                txaio.add_callbacks(d, success, _error)

            elif isinstance(msg, message.Event):
                if msg.subscription in self._subscriptions:
                    # fire all event handlers on subscription ..
                    for subscription in self._subscriptions[msg.subscription]:
                        handler = subscription.handler
                        topic = msg.topic or subscription.topic

                        if msg.enc_algo:
                            # FIXME: behavior in error cases (no keyring, decrypt issues, URI mismatch, ..)
                            if not self._payload_codec:
                                self.log.warn(
                                    "received encoded payload with enc_algo={enc_algo}, but no payload codec active - ignoring encoded payload!",
                                    enc_algo=msg.enc_algo,
                                )
                                return
                            else:
                                try:
                                    encoded_payload = EncodedPayload(
                                        msg.payload,
                                        msg.enc_algo,
                                        msg.enc_serializer,
                                        msg.enc_key,
                                    )
                                    decoded_topic, msg.args, msg.kwargs = (
                                        self._payload_codec.decode(
                                            False, topic, encoded_payload
                                        )
                                    )
                                except Exception as e:
                                    self.log.warn(
                                        "failed to decode application payload encoded with enc_algo={enc_algo}: {error}",
                                        error=e,
                                        enc_algo=msg.enc_algo,
                                    )
                                    return
                                else:
                                    if topic != decoded_topic:
                                        self.log.warn(
                                            "envelope topic URI does not match encoded one"
                                        )
                                        return

                        invoke_args = (handler.obj,) if handler.obj else tuple()
                        if msg.args:
                            invoke_args = invoke_args + tuple(msg.args)
                        invoke_kwargs = msg.kwargs if msg.kwargs else dict()

                        if handler.details_arg:
                            invoke_kwargs[handler.details_arg] = types.EventDetails(
                                subscription,
                                msg.publication,
                                publisher=msg.publisher,
                                publisher_authid=msg.publisher_authid,
                                publisher_authrole=msg.publisher_authrole,
                                topic=topic,
                                transaction_hash=msg.transaction_hash,
                                retained=msg.retained,
                                enc_algo=msg.enc_algo,
                                forward_for=msg.forward_for,
                            )

                        # FIXME: https://github.com/crossbario/autobahn-python/issues/764
                        def _success(_):
                            # Acknowledged Events -- only if we got the details header and
                            # the broker advertised it
                            if (
                                msg.x_acknowledged_delivery
                                and self._router_roles[
                                    "broker"
                                ].x_acknowledged_event_delivery
                            ):
                                if self._transport:
                                    response = message.EventReceived(msg.publication)
                                    self._transport.send(response)
                                else:
                                    self.log.warn(
                                        "successfully processed event with acknowledged delivery, but could not send ACK, since the transport was lost in the meantime"
                                    )

                        def _error(e):
                            errmsg = f"While firing {handler.fn} subscribed under {msg.subscription}."
                            return self._swallow_error(e, errmsg)

                        future = txaio.as_future(
                            handler.fn, *invoke_args, **invoke_kwargs
                        )
                        txaio.add_callbacks(future, _success, _error)

                else:
                    raise ProtocolError(
                        f"EVENT received for non-subscribed subscription ID {msg.subscription}"
                    )

            elif isinstance(msg, message.Published):
                if msg.request in self._publish_reqs:
                    # get and pop outstanding publish request
                    publish_request = self._publish_reqs.pop(msg.request)

                    if txaio.is_future(publish_request.on_reply) and txaio.is_called(
                        publish_request.on_reply
                    ):
                        return

                    # create a new publication object
                    publication = Publication(
                        msg.publication, was_encrypted=publish_request.was_encrypted
                    )

                    # resolve deferred/future for publishing successfully
                    txaio.resolve(publish_request.on_reply, publication)
                else:
                    raise ProtocolError(
                        f"PUBLISHED received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Subscribed):
                if msg.request in self._subscribe_reqs:
                    # get and pop outstanding subscribe request
                    request = self._subscribe_reqs.pop(msg.request)

                    if txaio.is_future(request.on_reply) and txaio.is_called(
                        request.on_reply
                    ):
                        return

                    # create new handler subscription list for subscription ID if not yet tracked
                    if msg.subscription not in self._subscriptions:
                        self._subscriptions[msg.subscription] = []

                    subscription = Subscription(
                        msg.subscription, request.topic, self, request.handler
                    )

                    # add handler to existing subscription
                    self._subscriptions[msg.subscription].append(subscription)

                    # resolve deferred/future for subscribing successfully
                    txaio.resolve(request.on_reply, subscription)
                else:
                    raise ProtocolError(
                        f"SUBSCRIBED received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Unsubscribed):
                if msg.request in self._unsubscribe_reqs:
                    # get and pop outstanding subscribe request
                    request = self._unsubscribe_reqs.pop(msg.request)

                    if txaio.is_future(request.on_reply) and txaio.is_called(
                        request.on_reply
                    ):
                        return

                    # if the subscription still exists, mark as inactive and remove ..
                    if request.subscription_id in self._subscriptions:
                        for subscription in self._subscriptions[
                            request.subscription_id
                        ]:
                            subscription.active = False
                        del self._subscriptions[request.subscription_id]

                    # resolve deferred/future for unsubscribing successfully
                    txaio.resolve(request.on_reply, 0)
                else:
                    raise ProtocolError(
                        f"UNSUBSCRIBED received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Result):
                if msg.request in self._call_reqs:
                    call_request = self._call_reqs[msg.request]
                    proc = call_request.procedure
                    enc_err = None

                    if msg.enc_algo:
                        if not self._payload_codec:
                            log_msg = (
                                "received encoded payload, but no payload codec active"
                            )
                            self.log.warn(log_msg)
                            enc_err = ApplicationError(
                                ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg
                            )
                        else:
                            try:
                                encoded_payload = EncodedPayload(
                                    msg.payload,
                                    msg.enc_algo,
                                    msg.enc_serializer,
                                    msg.enc_key,
                                )
                                decrypted_proc, msg.args, msg.kwargs = (
                                    self._payload_codec.decode(
                                        True, proc, encoded_payload
                                    )
                                )
                            except Exception as e:
                                self.log.warn(
                                    "failed to decrypt application payload 1: {err}",
                                    err=e,
                                )
                                enc_err = ApplicationError(
                                    ApplicationError.ENC_DECRYPT_ERROR,
                                    f"failed to decrypt application payload 1: {e}",
                                )
                            else:
                                if proc != decrypted_proc:
                                    self.log.warn(
                                        "URI within encrypted payload ('{decrypted_proc}') does not match the envelope ('{proc}')",
                                        decrypted_proc=decrypted_proc,
                                        proc=proc,
                                    )
                                    enc_err = ApplicationError(
                                        ApplicationError.ENC_TRUSTED_URI_MISMATCH,
                                        f"URI within encrypted payload ('{decrypted_proc}') does not match the envelope ('{proc}')",
                                    )

                    if msg.progress:
                        # process progressive call result

                        if call_request.options.on_progress:
                            if enc_err:
                                self.onUserError(
                                    enc_err,
                                    "could not deliver progressive call result, because payload decryption failed",
                                )
                            else:
                                kw = msg.kwargs or dict()
                                args = msg.args or tuple()

                                def _error(fail):
                                    self.onUserError(fail, "While firing on_progress")

                                if (
                                    call_request.options
                                    and call_request.options.details
                                ):
                                    prog_d = txaio.as_future(
                                        call_request.options.on_progress,
                                        types.CallResult(
                                            *msg.args,
                                            callee=msg.callee,
                                            callee_authid=msg.callee_authid,
                                            callee_authrole=msg.callee_authrole,
                                            forward_for=msg.forward_for,
                                            **msg.kwargs,
                                        ),
                                    )
                                else:
                                    prog_d = txaio.as_future(
                                        call_request.options.on_progress, *args, **kw
                                    )

                                txaio.add_callbacks(prog_d, None, _error)

                    else:
                        # process final call result

                        # drop original request
                        del self._call_reqs[msg.request]

                        # user callback that gets fired
                        on_reply = call_request.on_reply

                        if txaio.is_future(on_reply) and txaio.is_called(on_reply):
                            return

                        # above might already have rejected, so we guard ..
                        if enc_err:
                            txaio.reject(on_reply, enc_err)
                        else:
                            if msg.kwargs or (
                                call_request.options and call_request.options.details
                            ):
                                kwargs = msg.kwargs or {}
                                if msg.args:
                                    res = types.CallResult(
                                        *msg.args,
                                        callee=msg.callee,
                                        callee_authid=msg.callee_authid,
                                        callee_authrole=msg.callee_authrole,
                                        forward_for=msg.forward_for,
                                        **kwargs,
                                    )
                                else:
                                    res = types.CallResult(
                                        callee=msg.callee,
                                        callee_authid=msg.callee_authid,
                                        callee_authrole=msg.callee_authrole,
                                        forward_for=msg.forward_for,
                                        **kwargs,
                                    )
                                txaio.resolve(on_reply, res)
                            else:
                                if msg.args:
                                    if len(msg.args) > 1:
                                        res = types.CallResult(*msg.args)
                                        txaio.resolve(on_reply, res)
                                    else:
                                        txaio.resolve(on_reply, msg.args[0])
                                else:
                                    txaio.resolve(on_reply, None)
                else:
                    raise ProtocolError(
                        f"RESULT received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Invocation):
                if msg.request in self._invocations:
                    raise ProtocolError(
                        f"INVOCATION received for request ID {msg.request} already invoked"
                    )

                else:
                    if msg.registration not in self._registrations:
                        raise ProtocolError(
                            f"INVOCATION received for non-registered registration ID {msg.registration}"
                        )

                    else:
                        registration = self._registrations[msg.registration]
                        endpoint = registration.endpoint
                        proc = msg.procedure or registration.procedure
                        enc_err = None

                        if msg.enc_algo:
                            if not self._payload_codec:
                                log_msg = "received encrypted INVOCATION payload, but no keyring active"
                                self.log.warn(log_msg)
                                enc_err = ApplicationError(
                                    ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg
                                )
                            else:
                                try:
                                    encoded_payload = EncodedPayload(
                                        msg.payload,
                                        msg.enc_algo,
                                        msg.enc_serializer,
                                        msg.enc_key,
                                    )
                                    decrypted_proc, msg.args, msg.kwargs = (
                                        self._payload_codec.decode(
                                            False, proc, encoded_payload
                                        )
                                    )
                                except Exception as e:
                                    self.log.warn(
                                        "failed to decrypt INVOCATION payload: {err}",
                                        err=e,
                                    )
                                    enc_err = ApplicationError(
                                        ApplicationError.ENC_DECRYPT_ERROR,
                                        f"failed to decrypt INVOCATION payload: {e}",
                                    )
                                else:
                                    if proc != decrypted_proc:
                                        self.log.warn(
                                            "URI within encrypted INVOCATION payload ('{decrypted_proc}') "
                                            "does not match the envelope ('{proc}')",
                                            decrypted_proc=decrypted_proc,
                                            proc=proc,
                                        )
                                        enc_err = ApplicationError(
                                            ApplicationError.ENC_TRUSTED_URI_MISMATCH,
                                            f"URI within encrypted INVOCATION payload ('{decrypted_proc}') does not match the envelope ('{proc}')",
                                        )

                        if enc_err:
                            # when there was a problem decrypting the INVOCATION payload, we obviously can't invoke
                            # the endpoint, but return and
                            reply = self._message_from_exception(
                                message.Invocation.MESSAGE_TYPE, msg.request, enc_err
                            )
                            self._transport.send(reply)

                        else:
                            if endpoint.obj is not None:
                                invoke_args = (endpoint.obj,)
                            else:
                                invoke_args = tuple()

                            if msg.args:
                                invoke_args = invoke_args + tuple(msg.args)

                            invoke_kwargs = msg.kwargs if msg.kwargs else dict()

                            if endpoint.details_arg:
                                if msg.receive_progress:

                                    def progress(*args, **kwargs):
                                        assert args is None or type(args) in (
                                            list,
                                            tuple,
                                        )
                                        assert kwargs is None or type(kwargs) == dict

                                        encoded_payload = None
                                        if msg.enc_algo:
                                            if not self._payload_codec:
                                                raise Exception(
                                                    "trying to send encrypted payload, but no keyring active"
                                                )
                                            encoded_payload = (
                                                self._payload_codec.encode(
                                                    False, proc, args, kwargs
                                                )
                                            )

                                        if encoded_payload:
                                            progress_msg = message.Yield(
                                                msg.request,
                                                payload=encoded_payload.payload,
                                                progress=True,
                                                enc_algo=encoded_payload.enc_algo,
                                                enc_key=encoded_payload.enc_key,
                                                enc_serializer=encoded_payload.enc_serializer,
                                            )
                                        else:
                                            progress_msg = message.Yield(
                                                msg.request,
                                                args=args,
                                                kwargs=kwargs,
                                                progress=True,
                                            )

                                        self._transport.send(progress_msg)

                                else:
                                    progress = None

                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(
                                    registration,
                                    progress=progress,
                                    caller=msg.caller,
                                    caller_authid=msg.caller_authid,
                                    caller_authrole=msg.caller_authrole,
                                    procedure=proc,
                                    transaction_hash=msg.transaction_hash,
                                    enc_algo=msg.enc_algo,
                                )

                            on_reply = txaio.as_future(
                                endpoint.fn, *invoke_args, **invoke_kwargs
                            )

                            def success(res):
                                del self._invocations[msg.request]

                                encoded_payload = None
                                if msg.enc_algo:
                                    if not self._payload_codec:
                                        log_msg = "trying to send encrypted payload, but no keyring active"
                                        self.log.warn(log_msg)
                                    else:
                                        try:
                                            if isinstance(res, types.CallResult):
                                                encoded_payload = (
                                                    self._payload_codec.encode(
                                                        False,
                                                        proc,
                                                        res.results,
                                                        res.kwresults,
                                                    )
                                                )
                                            else:
                                                encoded_payload = (
                                                    self._payload_codec.encode(
                                                        False, proc, [res]
                                                    )
                                                )
                                        except Exception as e:
                                            self.log.warn(
                                                "failed to encrypt application payload: {err}",
                                                err=e,
                                            )

                                if encoded_payload:
                                    if isinstance(res, types.CallResult):
                                        reply = message.Yield(
                                            msg.request,
                                            payload=encoded_payload.payload,
                                            enc_algo=encoded_payload.enc_algo,
                                            enc_key=encoded_payload.enc_key,
                                            enc_serializer=encoded_payload.enc_serializer,
                                            callee=res.callee,
                                            callee_authid=res.callee_authid,
                                            callee_authrole=res.callee_authrole,
                                            forward_for=res.forward_for,
                                        )
                                    else:
                                        reply = message.Yield(
                                            msg.request,
                                            payload=encoded_payload.payload,
                                            enc_algo=encoded_payload.enc_algo,
                                            enc_key=encoded_payload.enc_key,
                                            enc_serializer=encoded_payload.enc_serializer,
                                        )
                                else:
                                    if isinstance(res, types.CallResult):
                                        reply = message.Yield(
                                            msg.request,
                                            args=res.results,
                                            kwargs=res.kwresults,
                                            callee=res.callee,
                                            callee_authid=res.callee_authid,
                                            callee_authrole=res.callee_authrole,
                                            forward_for=res.forward_for,
                                        )
                                    else:
                                        reply = message.Yield(msg.request, args=[res])

                                if self._transport is None:
                                    self.log.debug(
                                        f'Skipping result of "{registration.procedure}", request {msg.request} because transport disconnected.'
                                    )
                                    return

                                try:
                                    self._transport.send(reply)
                                except SerializationError as e:
                                    # the application-level payload returned from the invoked procedure can't be serialized
                                    error_reply = message.Error(
                                        message.Invocation.MESSAGE_TYPE,
                                        msg.request,
                                        ApplicationError.INVALID_PAYLOAD,
                                        args=[
                                            f'success return value (args={reply.args}, kwargs={reply.kwargs}) from invoked procedure "{registration.procedure}" could not be serialized: {e}'
                                        ],
                                    )
                                    self._transport.send(error_reply)
                                except PayloadExceededError as e:
                                    # the application-level payload returned from the invoked procedure, when serialized and framed
                                    # for the transport, exceeds the transport message/frame size limit
                                    error_reply = message.Error(
                                        message.Invocation.MESSAGE_TYPE,
                                        msg.request,
                                        ApplicationError.PAYLOAD_SIZE_EXCEEDED,
                                        args=[
                                            f'success return value (args={reply.args}, kwargs={reply.kwargs}) from invoked procedure "{registration.procedure}" exceeds transport size limit: {e}'
                                        ],
                                    )
                                    self._transport.send(error_reply)

                            def error(err):
                                del self._invocations[msg.request]

                                errmsg = txaio.failure_message(err)

                                try:
                                    self.onUserError(err, errmsg)
                                except:
                                    pass

                                formatted_tb = None
                                if self.traceback_app:
                                    formatted_tb = txaio.failure_format_traceback(err)

                                reply = self._message_from_exception(
                                    message.Invocation.MESSAGE_TYPE,
                                    msg.request,
                                    err.value,
                                    formatted_tb,
                                    msg.enc_algo,
                                )

                                try:
                                    self._transport.send(reply)
                                except SerializationError as e:
                                    # the application-level payload returned from the invoked procedure can't be serialized
                                    reply = message.Error(
                                        message.Invocation.MESSAGE_TYPE,
                                        msg.request,
                                        ApplicationError.INVALID_PAYLOAD,
                                        args=[
                                            f'error return value from invoked procedure "{registration.procedure}" could not be serialized: {e}'
                                        ],
                                    )
                                    self._transport.send(reply)
                                except PayloadExceededError as e:
                                    # the application-level payload returned from the invoked procedure, when serialized and framed
                                    # for the transport, exceeds the transport message/frame size limit
                                    reply = message.Error(
                                        message.Invocation.MESSAGE_TYPE,
                                        msg.request,
                                        ApplicationError.PAYLOAD_SIZE_EXCEEDED,
                                        args=[
                                            f'success return value from invoked procedure "{registration.procedure}" exceeds transport size limit: {e}'
                                        ],
                                    )
                                    self._transport.send(reply)

                                # we have handled the error, so we eat it
                                return None

                            self._invocations[msg.request] = InvocationRequest(
                                msg.request, on_reply
                            )

                            txaio.add_callbacks(on_reply, success, error)

            elif isinstance(msg, message.Interrupt):
                if msg.request not in self._invocations:
                    # raise ProtocolError("INTERRUPT received for non-pending invocation {0}".format(msg.request))
                    self.log.debug(
                        "INTERRUPT received for non-pending invocation {request}",
                        request=msg.request,
                    )
                else:
                    invoked = self._invocations[msg.request]
                    # this will result in a CancelledError which will
                    # be captured by the error handler around line 979
                    # to delete the invocation..
                    txaio.cancel(invoked.on_reply)

            elif isinstance(msg, message.Registered):
                if msg.request in self._register_reqs:
                    # get and pop outstanding register request
                    request = self._register_reqs.pop(msg.request)

                    if txaio.is_future(request.on_reply) and txaio.is_called(
                        request.on_reply
                    ):
                        return

                    # create new registration if not yet tracked
                    if msg.registration not in self._registrations:
                        registration = Registration(
                            self, msg.registration, request.procedure, request.endpoint
                        )
                        self._registrations[msg.registration] = registration
                    else:
                        raise ProtocolError(
                            f"REGISTERED received for already existing registration ID {msg.registration}"
                        )

                    txaio.resolve(request.on_reply, registration)
                else:
                    raise ProtocolError(
                        f"REGISTERED received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Unregistered):
                if msg.request == 0:
                    # this is a forced un-register either from a call
                    # to the wamp.* meta-api or the force_reregister
                    # option
                    try:
                        reg = self._registrations[msg.registration]
                    except KeyError:
                        raise ProtocolError(
                            "UNREGISTERED received for non-existant registration"
                            f" ID {msg.registration}"
                        )
                    self.log.debug(
                        "Router unregistered procedure '{proc}' with ID {id}",
                        proc=reg.procedure,
                        id=msg.registration,
                    )
                elif msg.request in self._unregister_reqs:
                    # get and pop outstanding subscribe request
                    request = self._unregister_reqs.pop(msg.request)

                    if txaio.is_future(request.on_reply) and txaio.is_called(
                        request.on_reply
                    ):
                        return

                    # if the registration still exists, mark as inactive and remove ..
                    if request.registration_id in self._registrations:
                        self._registrations[request.registration_id].active = False
                        del self._registrations[request.registration_id]

                    # resolve deferred/future for unregistering successfully
                    txaio.resolve(request.on_reply)
                else:
                    raise ProtocolError(
                        f"UNREGISTERED received for non-pending request ID {msg.request}"
                    )

            elif isinstance(msg, message.Error):
                # remove outstanding request and get the reply deferred/future
                on_reply = None

                # ERROR reply to CALL
                if (
                    msg.request_type == message.Call.MESSAGE_TYPE
                    and msg.request in self._call_reqs
                ):
                    on_reply = self._call_reqs.pop(msg.request).on_reply

                # ERROR reply to PUBLISH
                elif (
                    msg.request_type == message.Publish.MESSAGE_TYPE
                    and msg.request in self._publish_reqs
                ):
                    on_reply = self._publish_reqs.pop(msg.request).on_reply

                # ERROR reply to SUBSCRIBE
                elif (
                    msg.request_type == message.Subscribe.MESSAGE_TYPE
                    and msg.request in self._subscribe_reqs
                ):
                    on_reply = self._subscribe_reqs.pop(msg.request).on_reply

                # ERROR reply to UNSUBSCRIBE
                elif (
                    msg.request_type == message.Unsubscribe.MESSAGE_TYPE
                    and msg.request in self._unsubscribe_reqs
                ):
                    on_reply = self._unsubscribe_reqs.pop(msg.request).on_reply

                # ERROR reply to REGISTER
                elif (
                    msg.request_type == message.Register.MESSAGE_TYPE
                    and msg.request in self._register_reqs
                ):
                    on_reply = self._register_reqs.pop(msg.request).on_reply

                # ERROR reply to UNREGISTER
                elif (
                    msg.request_type == message.Unregister.MESSAGE_TYPE
                    and msg.request in self._unregister_reqs
                ):
                    on_reply = self._unregister_reqs.pop(msg.request).on_reply

                if on_reply:
                    if not txaio.is_called(on_reply):
                        txaio.reject(on_reply, self._exception_from_message(msg))
                else:
                    raise ProtocolError(
                        f"WampAppSession.onMessage(): ERROR received for non-pending request_type {msg.request_type} and request ID {msg.request}"
                    )

            else:
                raise ProtocolError(f"Unexpected message {msg.__class__}")

    @public
    def onClose(self, wasClean):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onClose`
        """
        self._transport = None

        if self._session_id:
            # fire callback and close the transport
            details = types.CloseDetails(
                reason=types.CloseDetails.REASON_TRANSPORT_LOST,
                message=f"WAMP transport was lost without closing the session {self._session_id} before",
            )
            d = txaio.as_future(self.onLeave, details)

            def success(arg):
                # XXX also: handle async
                self.fire("leave", self, details)
                return arg

            def _error(e):
                return self._swallow_error(e, "While firing onLeave")

            txaio.add_callbacks(d, success, _error)

            self._session_id = None

        d = txaio.as_future(self.onDisconnect)

        def success(arg):
            # XXX do we care about returning 'arg' properly?
            return self.fire("disconnect", self, was_clean=wasClean)

        def _error(e):
            return self._swallow_error(e, "While firing onDisconnect")

        txaio.add_callbacks(d, success, _error)

    @public
    def onChallenge(self, challenge: Challenge) -> str:
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onChallenge`
        """
        # Note: if we use NotImplementedError in below, the type checking in PyCharm
        # will recognize that, and complain about "not implemented abstract method"
        # in e.g. autobahn.twisted.wamp.ApplicationSession even though the latter
        # derives from this base class! IOW: don't change it;)
        raise RuntimeError(
            "received authentication challenge, but onChallenge not implemented"
        )

    @public
    def onJoin(self, details: SessionDetails):
        """
        Implements :meth:`autobahn.wamp.interfaces.ISession.onJoin`
        """

    @public
    def onWelcome(self, welcome: message.Welcome) -> str | None:
        """
        Implements :meth:`autobahn.wamp.interfaces.ISession.onWelcome`
        """

    def _errback_outstanding_requests(self, exc):
        """
        Errback any still outstanding requests with exc.
        """
        d = txaio.create_future_success(None)
        all_requests = [
            self._publish_reqs,
            self._subscribe_reqs,
            self._unsubscribe_reqs,
            self._call_reqs,
            self._register_reqs,
            self._unregister_reqs,
        ]
        outstanding = []
        for requests in all_requests:
            outstanding.extend(requests.values())
            requests.clear()

        if outstanding:
            self.log.info(
                "Cancelling {count} outstanding requests",
                count=len(outstanding),
            )
        for request in outstanding:
            self.log.debug(
                "cleaning up outstanding {request_type} request {request_id}, "
                "firing errback on user handler {request_on_reply}",
                request_on_reply=request.on_reply,
                request_id=request.request_id,
                request_type=request.__class__.__name__,
            )
            if not txaio.is_called(request.on_reply):
                txaio.reject(request.on_reply, exc)

            # wait for any async-ness in the error handlers for on_reply
            txaio.add_callbacks(
                d, lambda _: request.on_reply, lambda _: request.on_reply
            )
        return d

    @public
    def onLeave(self, details: CloseDetails):
        """
        Implements :meth:`autobahn.wamp.interfaces.ISession.onLeave`
        """
        if details.reason != CloseDetails.REASON_DEFAULT:
            self.log.warn(
                "session closed with reason {reason} [{message}]",
                reason=details.reason,
                message=details.message,
            )

        # fire ApplicationError on any currently outstanding requests
        exc = ApplicationError(details.reason, details.message)
        d = self._errback_outstanding_requests(exc)

        def disconnect(_):
            if self._transport:
                self.disconnect()

        txaio.add_callbacks(d, disconnect, disconnect)
        return d

    @public
    def leave(self, reason: str | None = None, message: str | None = None):
        """
        Implements :meth:`autobahn.wamp.interfaces.ISession.leave`
        """
        if not self._session_id:
            self.log.warn("session is not joined on a realm - no session to leave")
            return

        if not self._goodbye_sent:
            if not reason:
                reason = "wamp.close.normal"
            msg = wamp.message.Goodbye(reason=reason, message=message)
            self._transport.send(msg)
            self._goodbye_sent = True
        else:
            self.log.warn(
                "session was already requested to leave - not sending GOODBYE again"
            )

        is_closed = self._transport is None or self._transport.is_closed

        return is_closed

    @public
    def onDisconnect(self):
        """
        Implements :meth:`autobahn.wamp.interfaces.ISession.onDisconnect`
        """
        # fire TransportLost on any _still_ outstanding requests
        # (these should have been already cleaned up in onLeave() - when
        # this actually has fired)
        exc = exception.TransportLost()
        self._errback_outstanding_requests(exc)

    # FIXME:
    # def publish(self, topic: str, *args: Optional[List[Any]], **kwargs: Optional[Dict[str, Any]]) -> Optional[Publication]:

    @public
    def publish(self, topic: str, *args, **kwargs) -> Publication | None:
        """
        Implements :meth:`autobahn.wamp.interfaces.IPublisher.publish`
        """
        assert type(topic) == str
        assert args is None or type(args) in (list, tuple)
        assert kwargs is None or type(kwargs) == dict

        message.check_or_raise_uri(
            topic,
            message=f"{self.__class__.__name__}.publish()",
            strict=False,
            allow_empty_components=False,
            allow_none=False,
        )

        options = kwargs.pop("options", None)
        if options and not isinstance(options, types.PublishOptions):
            raise Exception("options must be of type a.w.t.PublishOptions")

        if not self._transport:
            raise exception.TransportLost()

        request_id = self._request_id_gen.next()

        encoded_payload = None
        if self._payload_codec:
            encoded_payload = self._payload_codec.encode(True, topic, args, kwargs)

        if encoded_payload:
            if options:
                msg = message.Publish(
                    request_id,
                    topic,
                    payload=encoded_payload.payload,
                    enc_algo=encoded_payload.enc_algo,
                    enc_key=encoded_payload.enc_key,
                    enc_serializer=encoded_payload.enc_serializer,
                    **options.message_attr(),
                )
            else:
                msg = message.Publish(
                    request_id,
                    topic,
                    payload=encoded_payload.payload,
                    enc_algo=encoded_payload.enc_algo,
                    enc_key=encoded_payload.enc_key,
                    enc_serializer=encoded_payload.enc_serializer,
                )
        else:
            if options:
                msg = message.Publish(
                    request_id,
                    topic,
                    args=args,
                    kwargs=kwargs,
                    **options.message_attr(),
                )
            else:
                msg = message.Publish(request_id, topic, args=args, kwargs=kwargs)

        if options:
            if options.correlation_id is not None:
                msg.correlation_id = options.correlation_id
            if options.correlation_uri is not None:
                msg.correlation_uri = options.correlation_uri
            if options.correlation_is_anchor is not None:
                msg.correlation_is_anchor = options.correlation_is_anchor
            if options.correlation_is_last is not None:
                msg.correlation_is_last = options.correlation_is_last

        if options and options.acknowledge:
            # only acknowledged publications expect a reply ..
            on_reply = txaio.create_future()
            self._publish_reqs[request_id] = PublishRequest(
                request_id, on_reply, was_encrypted=(encoded_payload is not None)
            )
        else:
            on_reply = None

        try:
            # Notes:
            #
            # * this might raise autobahn.wamp.exception.SerializationError
            #   when the user payload cannot be serialized
            # * we have to setup a PublishRequest() in _publish_reqs _before_
            #   calling transpor.send(), because a mock- or side-by-side transport
            #   will immediately lead on an incoming WAMP message in onMessage()
            #
            self._transport.send(msg)
        except Exception as e:
            if request_id in self._publish_reqs:
                del self._publish_reqs[request_id]
            raise e

        return on_reply

    @public
    def subscribe(
        self,
        handler: Callable | Any,
        topic: str | None = None,
        options: SubscribeOptions | None = None,
        check_types: bool | None = None,
    ) -> Subscription | list[Subscription]:
        """
        Implements :meth:`autobahn.wamp.interfaces.ISubscriber.subscribe`
        """
        assert (callable(handler) and topic is not None) or (
            hasattr(handler, "__class__") and not check_types
        )
        assert topic is None or type(topic) == str
        assert options is None or isinstance(options, types.SubscribeOptions)

        if not self._transport:
            raise exception.TransportLost()

        def _subscribe(obj, fn, topic, options, check_types):
            message.check_or_raise_uri(
                topic,
                message=f"{self.__class__.__name__}.subscribe()",
                strict=False,
                allow_empty_components=True,
                allow_none=False,
            )

            request_id = self._request_id_gen.next()
            on_reply = txaio.create_future()
            if check_types:
                fn = self.type_check(fn)
            handler_obj = Handler(fn, obj, options.details_arg if options else None)
            self._subscribe_reqs[request_id] = SubscribeRequest(
                request_id, topic, on_reply, handler_obj
            )

            if options:
                msg = message.Subscribe(request_id, topic, **options.message_attr())
            else:
                msg = message.Subscribe(request_id, topic)

            if options:
                if options.correlation_id is not None:
                    msg.correlation_id = options.correlation_id
                if options.correlation_uri is not None:
                    msg.correlation_uri = options.correlation_uri
                if options.correlation_is_anchor is not None:
                    msg.correlation_is_anchor = options.correlation_is_anchor
                if options.correlation_is_last is not None:
                    msg.correlation_is_last = options.correlation_is_last

            self._transport.send(msg)
            return on_reply

        if callable(handler):
            # subscribe a single handler
            return _subscribe(None, handler, topic, options, check_types)

        else:
            # subscribe all methods on an object decorated with "wamp.subscribe"
            on_replies = []
            for k in inspect.getmembers(handler.__class__, is_method_or_function):
                proc = k[1]
                if "_wampuris" in proc.__dict__:
                    for pat in proc.__dict__["_wampuris"]:
                        if pat.is_handler():
                            _uri = pat.uri()
                            subopts = pat.options or options
                            if subopts is None:
                                if pat.uri_type == uri.Pattern.URI_TYPE_WILDCARD:
                                    subopts = types.SubscribeOptions(match="wildcard")
                                else:
                                    subopts = types.SubscribeOptions(match="exact")
                            on_replies.append(
                                _subscribe(
                                    handler, proc, _uri, subopts, pat._check_types
                                )
                            )

            return txaio.gather(on_replies, consume_exceptions=True)

    def _unsubscribe(self, subscription):
        """
        Called from :meth:`autobahn.wamp.protocol.Subscription.unsubscribe`
        """
        assert isinstance(subscription, Subscription)
        assert subscription.active
        assert subscription.id in self._subscriptions
        assert subscription in self._subscriptions[subscription.id]

        if not self._transport:
            raise exception.TransportLost()

        # remove handler subscription and mark as inactive
        self._subscriptions[subscription.id].remove(subscription)
        subscription.active = False

        # number of handler subscriptions left ..
        scount = len(self._subscriptions[subscription.id])

        if scount == 0:
            # if the last handler was removed, unsubscribe from broker ..
            request_id = self._request_id_gen.next()

            on_reply = txaio.create_future()
            self._unsubscribe_reqs[request_id] = UnsubscribeRequest(
                request_id, on_reply, subscription.id
            )

            msg = message.Unsubscribe(request_id, subscription.id)

            self._transport.send(msg)
            return on_reply
        else:
            # there are still handlers active on the subscription!
            return txaio.create_future_success(scount)

    @public
    def call(self, procedure: str, *args, **kwargs) -> Any | CallResult:
        """
        Implements :meth:`autobahn.wamp.interfaces.ICaller.call`

        .. note::

            Regarding type hints for ``*args`` and ``**kwargs``, doesn't work as we
            can receive any Python types as list items or dict values, and because
            of what is discussed here
            https://adamj.eu/tech/2021/05/11/python-type-hints-args-and-kwargs/
        """
        assert type(procedure) == str
        assert args is None or type(args) in (list, tuple)
        assert kwargs is None or type(kwargs) == dict

        message.check_or_raise_uri(
            procedure,
            message=f"{self.__class__.__name__}.call()",
            strict=False,
            allow_empty_components=False,
            allow_none=False,
        )

        options = kwargs.pop("options", None)
        if options and not isinstance(options, types.CallOptions):
            raise Exception("options must be of type a.w.t.CallOptions")

        if not self._transport:
            raise exception.TransportLost()

        request_id = self._request_id_gen.next()

        encoded_payload = None
        if self._payload_codec:
            try:
                encoded_payload = self._payload_codec.encode(
                    True, procedure, args, kwargs
                )
            except:
                self.log.failure()
                raise

        if encoded_payload:
            if options:
                msg = message.Call(
                    request_id,
                    procedure,
                    payload=encoded_payload.payload,
                    enc_algo=encoded_payload.enc_algo,
                    enc_key=encoded_payload.enc_key,
                    enc_serializer=encoded_payload.enc_serializer,
                    **options.message_attr(),
                )
            else:
                msg = message.Call(
                    request_id,
                    procedure,
                    payload=encoded_payload.payload,
                    enc_algo=encoded_payload.enc_algo,
                    enc_key=encoded_payload.enc_key,
                    enc_serializer=encoded_payload.enc_serializer,
                )
        else:
            if options:
                msg = message.Call(
                    request_id,
                    procedure,
                    args=args,
                    kwargs=kwargs,
                    **options.message_attr(),
                )
            else:
                msg = message.Call(request_id, procedure, args=args, kwargs=kwargs)

        if options:
            if options.correlation_id is not None:
                msg.correlation_id = options.correlation_id
            if options.correlation_uri is not None:
                msg.correlation_uri = options.correlation_uri
            if options.correlation_is_anchor is not None:
                msg.correlation_is_anchor = options.correlation_is_anchor
            if options.correlation_is_last is not None:
                msg.correlation_is_last = options.correlation_is_last

        def canceller(d):
            cancel_msg = message.Cancel(request_id)
            self._transport.send(cancel_msg)
            # since we announced support for cancelling, we should
            # definitely get an Error back for our Cancel which will
            # clean up this invocation

        on_reply = txaio.create_future(canceller=canceller)
        self._call_reqs[request_id] = CallRequest(
            request_id, procedure, on_reply, options
        )

        try:
            # Notes:
            #
            # * this might raise autobahn.wamp.exception.SerializationError
            #   when the user payload cannot be serialized
            # * we have to setup a CallRequest() in _call_reqs _before_
            #   calling transpor.send(), because a mock- or side-by-side transport
            #   will immediately lead on an incoming WAMP message in onMessage()
            #
            self._transport.send(msg)
        except:
            if request_id in self._call_reqs:
                del self._call_reqs[request_id]
            raise

        return on_reply

    @public
    def register(
        self,
        endpoint: Callable | Any,
        procedure: str | None = None,
        options: RegisterOptions | None = None,
        prefix: str | None = None,
        check_types: bool | None = None,
    ) -> Registration | list[Registration]:
        """
        Implements :meth:`autobahn.wamp.interfaces.ICallee.register`
        """
        assert (callable(endpoint) and procedure is not None) or (
            hasattr(endpoint, "__class__") and not check_types
        )

        if not self._transport:
            raise exception.TransportLost()

        def _register(obj, fn, procedure, options, check_types):
            message.check_or_raise_uri(
                procedure,
                message=f"{self.__class__.__name__}.register()",
                strict=False,
                allow_empty_components=True,
                allow_none=False,
            )

            request_id = self._request_id_gen.next()
            on_reply = txaio.create_future()
            if check_types:
                fn = self.type_check(fn)
            endpoint_obj = Endpoint(fn, obj, options.details_arg if options else None)
            if prefix is not None:
                procedure = f"{prefix}{procedure}"
            self._register_reqs[request_id] = RegisterRequest(
                request_id, on_reply, procedure, endpoint_obj
            )

            if options:
                msg = message.Register(request_id, procedure, **options.message_attr())
            else:
                msg = message.Register(request_id, procedure)

            if options:
                if options.correlation_id is not None:
                    msg.correlation_id = options.correlation_id
                if options.correlation_uri is not None:
                    msg.correlation_uri = options.correlation_uri
                if options.correlation_is_anchor is not None:
                    msg.correlation_is_anchor = options.correlation_is_anchor
                if options.correlation_is_last is not None:
                    msg.correlation_is_last = options.correlation_is_last

            self._transport.send(msg)
            return on_reply

        if callable(endpoint):
            # register a single callable
            return _register(None, endpoint, procedure, options, check_types)

        else:
            # register all methods on an object decorated with "wamp.register"
            on_replies = []
            for k in inspect.getmembers(endpoint.__class__, is_method_or_function):
                proc = k[1]
                if "_wampuris" in proc.__dict__:
                    for pat in proc.__dict__["_wampuris"]:
                        if pat.is_endpoint():
                            _uri = pat.uri()
                            regopts = pat.options or options
                            on_replies.append(
                                _register(
                                    endpoint, proc, _uri, regopts, pat._check_types
                                )
                            )

            return txaio.gather(on_replies, consume_exceptions=True)

    def _unregister(self, registration):
        """
        Called from :meth:`autobahn.wamp.protocol.Registration.unregister`
        """
        assert isinstance(registration, Registration)
        assert registration.active
        assert registration.id in self._registrations

        if not self._transport:
            raise exception.TransportLost()

        request_id = self._request_id_gen.next()

        on_reply = txaio.create_future()
        self._unregister_reqs[request_id] = UnregisterRequest(
            request_id, on_reply, registration.id
        )

        msg = message.Unregister(request_id, registration.id)

        self._transport.send(msg)
        return on_reply


class _SessionShim(ApplicationSession):
    """
    shim that lets us present pep8 API for user-classes to override,
    but also backwards-compatible for existing code using
    ApplicationSession "directly".

    **NOTE:** this is not public or intended for use; you should import
    either autobahn.asyncio.wamp.Session or
    autobahn.twisted.wamp.Session depending on which async
    framework yo're using.
    """

    #: name -> IAuthenticator
    _authenticators = None

    def onJoin(self, details):
        return self.on_join(details)

    def onConnect(self):
        if self._authenticators:
            # authid, authrole *must* match across all authenticators
            # (checked in add_authenticator) so these lists are either
            # [None] or [None, 'some_authid']
            authid = [
                x._args.get("authid", None) for x in self._authenticators.values()
            ][-1]
            authrole = [
                x._args.get("authrole", None) for x in self._authenticators.values()
            ][-1]
            # we need a "merged" authextra here because we can send a
            # list of acceptable authmethods, but only a single
            # authextra dict
            authextra = self._merged_authextra()
            self.join(
                self.config.realm,
                authmethods=list(self._authenticators.keys()),
                authid=authid,
                authrole=authrole,
                authextra=authextra,
            )
        else:
            self.on_connect()

    def onChallenge(self, challenge):
        try:
            authenticator = self._authenticators[challenge.method]
        except KeyError:
            raise RuntimeError(
                f"Received challenge for unknown authmethod '{challenge.method}' [authenticators={str(sorted(self._authenticators.keys()))}]"
            )
        return authenticator.on_challenge(self, challenge)

    def onWelcome(self, msg):
        if msg.authmethod is None or self._authenticators is None:
            # no authentication
            return
        try:
            authenticator = self._authenticators[msg.authmethod]
        except KeyError:
            raise RuntimeError(
                f"Received onWelcome for unknown authmethod '{msg.authmethod}' [authenticators={str(sorted(self._authenticators.keys()))}]"
            )
        return authenticator.on_welcome(self, msg.authextra)

    def onLeave(self, details):
        return self.on_leave(details)

    def onDisconnect(self):
        return self.on_disconnect()

    # experimental authentication API

    def add_authenticator(self, authenticator):
        assert isinstance(authenticator, IAuthenticator)
        if self._authenticators is None:
            self._authenticators = {}

        # before adding this authenticator we need to validate that
        # it's consistent with any other authenticators we may have --
        # for example, they must all agree on "authid" etc because
        # .join() only takes one value for all of those.

        def at_most_one(name):
            uni = set(
                [
                    a._args[name]
                    for a in list(self._authenticators.values()) + [authenticator]
                    if name in a._args
                ]
            )
            if len(uni) > 1:
                raise ValueError(
                    "Inconsistent {}s: {}".format(
                        name,
                        " ".join(uni),
                    )
                )

        # all authids must match
        at_most_one("authid")

        # all authroles must match
        at_most_one("authrole")

        # can we do anything else other than merge all authextra keys?
        # here we check that any duplicate keys have the same values
        authextra = authenticator.authextra
        merged = self._merged_authextra()
        for k, v in merged.items():
            if k in authextra and authextra[k] != v:
                raise ValueError(
                    f"Inconsistent authextra values for '{k}': '{v}' vs '{authextra[k]}'"
                )

        # validation complete, add it
        self._authenticators[authenticator.name] = authenticator

    def _merged_authextra(self):
        """
        internal helper

        :returns: a single 'authextra' dict, consisting of all keys
            from any authenticator's authextra.

        Note that when the authenticator was added, we already checked
        that any keys it does contain has the same value as any
        existing authextra.
        """
        authextras = [a.authextra for a in self._authenticators.values()]

        def extract_keys(x, y):
            return x | set(y.keys())

        unique_keys = reduce(extract_keys, authextras, set())

        def first_value_for(k):
            """
            for anything already in self._authenticators, we checked
            that it has the same value for any keys in its authextra --
            so here we just extract the first one
            """
            for authextra in authextras:
                if k in authextra:
                    return authextra[k]
            # "can't" happen
            raise ValueError(f"No values for '{k}'")

        return {k: first_value_for(k) for k in unique_keys}

    # these are the actual "new API" methods (i.e. snake_case)
    #

    def on_join(self, details):
        pass

    def on_leave(self, details):
        self.disconnect()

    def on_connect(self):
        self.join(self.config.realm)

    def on_disconnect(self):
        pass


# ISession.register collides with the abc.ABCMeta.register method
ISession.abc_register(ApplicationSession)


class ApplicationSessionFactory:
    """
    WAMP endpoint session factory.
    """

    session = ApplicationSession
    """
    WAMP application session class to be used in this factory.
    """

    def __init__(self, config=None):
        """

        :param config: The default component configuration.
        :type config: instance of :class:`autobahn.wamp.types.ComponentConfig`
        """
        self.config = config or types.ComponentConfig(realm="realm1")

    def __call__(self):
        """
        Creates a new WAMP application session.

        :returns: -- An instance of the WAMP application session class as
                     given by `self.session`.
        """
        session = self.session(self.config)
        session.factory = self
        return session



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class HelloNew(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = HelloNew()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsHelloNew(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # HelloNew
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # HelloNew
    def Roles(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.ClientRoles import ClientRoles

            obj = ClientRoles()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # HelloNew
    def Realm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # HelloNew
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # HelloNew
    def Authrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # HelloNew
    def Authmode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # HelloNew
    def Authfactor1Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # HelloNew
    def Authfactor1(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            from autobahn.flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # HelloNew
    def Authfactor2Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # HelloNew
    def Authfactor2(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            from autobahn.flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # HelloNew
    def Authfactor3Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # HelloNew
    def Authfactor3(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            from autobahn.flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # HelloNew
    def Resumable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # HelloNew
    def ResumeSession(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # HelloNew
    def ResumeToken(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def HelloNewStart(builder):
    builder.StartObject(14)


def Start(builder):
    HelloNewStart(builder)


def HelloNewAddRoles(builder, roles):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(roles), 0
    )


def AddRoles(builder, roles):
    HelloNewAddRoles(builder, roles)


def HelloNewAddRealm(builder, realm):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(realm), 0
    )


def AddRealm(builder, realm):
    HelloNewAddRealm(builder, realm)


def HelloNewAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0
    )


def AddAuthid(builder, authid):
    HelloNewAddAuthid(builder, authid)


def HelloNewAddAuthrole(builder, authrole):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(authrole), 0
    )


def AddAuthrole(builder, authrole):
    HelloNewAddAuthrole(builder, authrole)


def HelloNewAddAuthmode(builder, authmode):
    builder.PrependUint8Slot(4, authmode, 0)


def AddAuthmode(builder, authmode):
    HelloNewAddAuthmode(builder, authmode)


def HelloNewAddAuthfactor1Type(builder, authfactor1Type):
    builder.PrependUint8Slot(5, authfactor1Type, 0)


def AddAuthfactor1Type(builder, authfactor1Type):
    HelloNewAddAuthfactor1Type(builder, authfactor1Type)


def HelloNewAddAuthfactor1(builder, authfactor1):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(authfactor1), 0
    )


def AddAuthfactor1(builder, authfactor1):
    HelloNewAddAuthfactor1(builder, authfactor1)


def HelloNewAddAuthfactor2Type(builder, authfactor2Type):
    builder.PrependUint8Slot(7, authfactor2Type, 0)


def AddAuthfactor2Type(builder, authfactor2Type):
    HelloNewAddAuthfactor2Type(builder, authfactor2Type)


def HelloNewAddAuthfactor2(builder, authfactor2):
    builder.PrependUOffsetTRelativeSlot(
        8, flatbuffers.number_types.UOffsetTFlags.py_type(authfactor2), 0
    )


def AddAuthfactor2(builder, authfactor2):
    HelloNewAddAuthfactor2(builder, authfactor2)


def HelloNewAddAuthfactor3Type(builder, authfactor3Type):
    builder.PrependUint8Slot(9, authfactor3Type, 0)


def AddAuthfactor3Type(builder, authfactor3Type):
    HelloNewAddAuthfactor3Type(builder, authfactor3Type)


def HelloNewAddAuthfactor3(builder, authfactor3):
    builder.PrependUOffsetTRelativeSlot(
        10, flatbuffers.number_types.UOffsetTFlags.py_type(authfactor3), 0
    )


def AddAuthfactor3(builder, authfactor3):
    HelloNewAddAuthfactor3(builder, authfactor3)


def HelloNewAddResumable(builder, resumable):
    builder.PrependBoolSlot(11, resumable, 0)


def AddResumable(builder, resumable):
    HelloNewAddResumable(builder, resumable)


def HelloNewAddResumeSession(builder, resumeSession):
    builder.PrependUint64Slot(12, resumeSession, 0)


def AddResumeSession(builder, resumeSession):
    HelloNewAddResumeSession(builder, resumeSession)


def HelloNewAddResumeToken(builder, resumeToken):
    builder.PrependUOffsetTRelativeSlot(
        13, flatbuffers.number_types.UOffsetTFlags.py_type(resumeToken), 0
    )


def AddResumeToken(builder, resumeToken):
    HelloNewAddResumeToken(builder, resumeToken)


def HelloNewEnd(builder):
    return builder.EndObject()


def End(builder):
    return HelloNewEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import asyncio
import signal
import ssl
from functools import wraps

import txaio

from autobahn.asyncio.rawsocket import WampRawSocketClientFactory
from autobahn.asyncio.wamp import Session
from autobahn.asyncio.websocket import WampWebSocketClientFactory
from autobahn.wamp import component
from autobahn.wamp.exception import TransportLost
from autobahn.wamp.serializer import (
    create_transport_serializer,
    create_transport_serializers,
)

__all__ = ("Component", "run")


def _unique_list(seq):
    """
    Return a list with unique elements from sequence, preserving order.
    """
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]


def _camel_case_from_snake_case(snake):
    parts = snake.split("_")
    return parts[0] + "".join(s.capitalize() for s in parts[1:])


def _create_transport_factory(loop, transport, session_factory):
    """
    Create a WAMP-over-XXX transport factory.
    """
    if transport.type == "websocket":
        serializers = create_transport_serializers(transport)
        factory = WampWebSocketClientFactory(
            session_factory,
            url=transport.url,
            serializers=serializers,
            proxy=transport.proxy,  # either None or a dict with host, port
        )

    elif transport.type == "rawsocket":
        serializer = create_transport_serializer(transport.serializers[0])
        factory = WampRawSocketClientFactory(session_factory, serializer=serializer)

    else:
        assert False, "should not arrive here"

    # set the options one at a time so we can give user better feedback
    for k, v in transport.options.items():
        try:
            factory.setProtocolOptions(**{k: v})
        except (TypeError, KeyError):
            # this allows us to document options as snake_case
            # until everything internally is upgraded from
            # camelCase
            try:
                factory.setProtocolOptions(**{_camel_case_from_snake_case(k): v})
            except (TypeError, KeyError):
                raise ValueError(
                    f"Unknown {transport.type} transport option: {k}={v}"
                )
    return factory


class Component(component.Component):
    """
    A component establishes a transport and attached a session
    to a realm using the transport for communication.

    The transports a component tries to use can be configured,
    as well as the auto-reconnect strategy.
    """

    log = txaio.make_logger()

    session_factory = Session
    """
    The factory of the session we will instantiate.
    """

    def _is_ssl_error(self, e):
        """
        Internal helper.
        """
        return isinstance(e, ssl.SSLError)

    def _check_native_endpoint(self, endpoint):
        if isinstance(endpoint, dict):
            if "tls" in endpoint:
                tls = endpoint["tls"]
                if isinstance(tls, dict | bool):
                    pass
                elif isinstance(tls, ssl.SSLContext):
                    pass
                else:
                    raise ValueError(
                        "'tls' configuration must be a dict, bool or "
                        "SSLContext instance"
                    )
        else:
            raise ValueError(
                "'endpoint' configuration must be a dict or IStreamClientEndpoint"
                " provider"
            )

    # async function
    def _connect_transport(self, loop, transport, session_factory, done):
        """
        Create and connect a WAMP-over-XXX transport.
        """
        factory = _create_transport_factory(loop, transport, session_factory)

        # XXX the rest of this should probably be factored into its
        # own method (or three!)...

        if transport.proxy:
            timeout = transport.endpoint.get("timeout", 10)  # in seconds
            if type(timeout) != int:
                raise ValueError(
                    f"invalid type {type(timeout)} for timeout in client endpoint configuration"
                )
            # do we support HTTPS proxies?

            f = loop.create_connection(
                protocol_factory=factory,
                host=transport.proxy["host"],
                port=transport.proxy["port"],
            )
            time_f = asyncio.ensure_future(asyncio.wait_for(f, timeout=timeout))
            return self._wrap_connection_future(transport, done, time_f)

        elif transport.endpoint["type"] == "tcp":
            version = transport.endpoint.get("version", 4)
            if version not in [4, 6]:
                raise ValueError(
                    f"invalid IP version {version} in client endpoint configuration"
                )

            host = transport.endpoint["host"]
            if type(host) != str:
                raise ValueError(
                    f"invalid type {type(host)} for host in client endpoint configuration"
                )

            port = transport.endpoint["port"]
            if type(port) != int:
                raise ValueError(
                    f"invalid type {type(port)} for port in client endpoint configuration"
                )

            timeout = transport.endpoint.get("timeout", 10)  # in seconds
            if type(timeout) != int:
                raise ValueError(
                    f"invalid type {type(timeout)} for timeout in client endpoint configuration"
                )

            tls = transport.endpoint.get("tls", None)
            tls_hostname = None

            # create a TLS enabled connecting TCP socket
            if tls:
                if isinstance(tls, dict):
                    for k in tls.keys():
                        if k not in ["hostname", "trust_root"]:
                            raise ValueError(
                                f"Invalid key '{k}' in 'tls' config"
                            )
                    hostname = tls.get("hostname", host)
                    if type(hostname) != str:
                        raise ValueError(
                            f"invalid type {hostname} for hostname in TLS client endpoint configuration"
                        )
                    cert_fname = tls.get("trust_root", None)

                    tls_hostname = hostname
                    tls = True
                    if cert_fname is not None:
                        tls = ssl.create_default_context(
                            purpose=ssl.Purpose.SERVER_AUTH,
                            cafile=cert_fname,
                        )

                elif isinstance(tls, ssl.SSLContext):
                    # tls=<an SSLContext> is valid
                    tls_hostname = host

                elif tls in [False, True]:
                    if tls:
                        tls_hostname = host

                else:
                    raise RuntimeError(
                        f'unknown type {type(tls)} for "tls" configuration in transport'
                    )

            f = loop.create_connection(
                protocol_factory=factory,
                host=host,
                port=port,
                ssl=tls,
                server_hostname=tls_hostname,
            )
            time_f = asyncio.ensure_future(asyncio.wait_for(f, timeout=timeout))
            return self._wrap_connection_future(transport, done, time_f)

        elif transport.endpoint["type"] == "unix":
            path = transport.endpoint["path"]
            timeout = int(transport.endpoint.get("timeout", 10))  # in seconds

            f = loop.create_unix_connection(
                protocol_factory=factory,
                path=path,
            )
            time_f = asyncio.ensure_future(asyncio.wait_for(f, timeout=timeout))
            return self._wrap_connection_future(transport, done, time_f)

        else:
            assert False, "should not arrive here"

    def _wrap_connection_future(self, transport, done, conn_f):
        def on_connect_success(result):
            # async connect call returns a 2-tuple
            transport, proto = result

            # in the case where we .abort() the transport / connection
            # during setup, we still get on_connect_success but our
            # transport is already closed (this will happen if
            # e.g. there's an "open handshake timeout") -- I don't
            # know if there's a "better" way to detect this? #python
            # doesn't know of one, anyway
            if transport.is_closing():
                if not txaio.is_called(done):
                    reason = getattr(
                        proto, "_onclose_reason", "Connection already closed"
                    )
                    txaio.reject(done, TransportLost(reason))
                return

            # if e.g. an SSL handshake fails, we will have
            # successfully connected (i.e. get here) but need to
            # 'listen' for the "connection_lost" from the underlying
            # protocol in case of handshake failure .. so we wrap
            # it. Also, we don't increment transport.success_count
            # here on purpose (because we might not succeed).

            # XXX double-check that asyncio behavior on TLS handshake
            # failures is in fact as described above
            orig = proto.connection_lost

            @wraps(orig)
            def lost(fail):
                rtn = orig(fail)
                if not txaio.is_called(done):
                    # asyncio will call connection_lost(None) in case of
                    # a transport failure, in which case we create an
                    # appropriate exception
                    if fail is None:
                        fail = TransportLost("failed to complete connection")
                    txaio.reject(done, fail)
                return rtn

            proto.connection_lost = lost

        def on_connect_failure(err):
            transport.connect_failures += 1
            # failed to establish a connection in the first place
            txaio.reject(done, err)

        txaio.add_callbacks(conn_f, on_connect_success, None)
        # the errback is added as a second step so it gets called if
        # there as an error in on_connect_success itself.
        txaio.add_callbacks(conn_f, None, on_connect_failure)
        return conn_f

    # async function
    def start(self, loop=None):
        """
        This starts the Component, which means it will start connecting
        (and re-connecting) to its configured transports. A Component
        runs until it is "done", which means one of:
        - There was a "main" function defined, and it completed successfully;
        - Something called ``.leave()`` on our session, and we left successfully;
        - ``.stop()`` was called, and completed successfully;
        - none of our transports were able to connect successfully (failure);

        :returns: a Future which will resolve (to ``None``) when we are
            "done" or with an error if something went wrong.
        """

        if loop is None:
            self.log.warn("Using default loop")
            loop = asyncio.get_event_loop()

        return self._start(loop=loop)


def run(components, start_loop=True, log_level="info"):
    """
    High-level API to run a series of components.

    This will only return once all the components have stopped
    (including, possibly, after all re-connections have failed if you
    have re-connections enabled). Under the hood, this calls

    XXX fixme for asyncio

    -- if you wish to manage the loop yourself, use the
    :meth:`autobahn.asyncio.component.Component.start` method to start
    each component yourself.

    :param components: the Component(s) you wish to run
    :type components: instance or list of :class:`autobahn.asyncio.component.Component`

    :param start_loop: When ``True`` (the default) this method
        start a new asyncio loop.
    :type start_loop: bool

    :param log_level: a valid log-level (or None to avoid calling start_logging)
    :type log_level: string
    """

    # actually, should we even let people "not start" the logging? I'm
    # not sure that's wise... (double-check: if they already called
    # txaio.start_logging() what happens if we call it again?)
    if log_level is not None:
        txaio.start_logging(level=log_level)
    loop = asyncio.get_event_loop()
    if loop.is_closed():
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = asyncio.get_event_loop()
        txaio.config.loop = loop
    log = txaio.make_logger()

    # see https://github.com/python/asyncio/issues/341 asyncio has
    # "odd" handling of KeyboardInterrupt when using Tasks (as
    # run_until_complete does). Another option is to just resture
    # default SIGINT handling, which is to exit:
    #   import signal
    #   signal.signal(signal.SIGINT, signal.SIG_DFL)

    async def nicely_exit(signal):
        log.info("Shutting down due to {signal}", signal=signal)

        try:
            tasks = asyncio.Task.all_tasks()
        except AttributeError:
            # this changed with python >= 3.7
            tasks = asyncio.all_tasks()

        for task in tasks:
            # Do not cancel the current task.

            try:
                current_task = asyncio.Task.current_task()
            except AttributeError:
                current_task = asyncio.current_task()

            if task is not current_task:
                task.cancel()

        def cancel_all_callback(fut):
            try:
                fut.result()
            except asyncio.CancelledError:
                log.debug("All task cancelled")
            except Exception as e:
                log.error("Error while shutting down: {exception}", exception=e)
            finally:
                loop.stop()

        fut = asyncio.gather(*tasks)
        fut.add_done_callback(cancel_all_callback)

    try:
        loop.add_signal_handler(
            signal.SIGINT, lambda: asyncio.ensure_future(nicely_exit("SIGINT"))
        )
        loop.add_signal_handler(
            signal.SIGTERM, lambda: asyncio.ensure_future(nicely_exit("SIGTERM"))
        )
    except NotImplementedError:
        # signals are not available on Windows
        pass

    def done_callback(loop, arg):
        loop.stop()

    # returns a future; could run_until_complete() but see below
    component._run(loop, components, done_callback)

    if start_loop:
        try:
            loop.run_forever()
            # this is probably more-correct, but then you always get
            # "Event loop stopped before Future completed":
            # loop.run_until_complete(f)
        except asyncio.CancelledError:
            pass
        # finally:
        #     signal.signal(signal.SIGINT, signal.SIG_DFL)
        #     signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # Close the event loop at the end, otherwise an exception is
        # thrown. https://bugs.python.org/issue23548
        loop.close()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class Register(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Register()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRegister(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Register
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Register
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Register
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Register
    def Procedure(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Register
    def Match(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Register
    def Invoke(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Register
    def Concurrency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # Register
    def ForceReregister(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def RegisterStart(builder):
    builder.StartObject(7)


def Start(builder):
    RegisterStart(builder)


def RegisterAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    RegisterAddSession(builder, session)


def RegisterAddRequest(builder, request):
    builder.PrependUint64Slot(1, request, 0)


def AddRequest(builder, request):
    RegisterAddRequest(builder, request)


def RegisterAddProcedure(builder, procedure):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(procedure), 0
    )


def AddProcedure(builder, procedure):
    RegisterAddProcedure(builder, procedure)


def RegisterAddMatch(builder, match):
    builder.PrependUint8Slot(3, match, 0)


def AddMatch(builder, match):
    RegisterAddMatch(builder, match)


def RegisterAddInvoke(builder, invoke):
    builder.PrependUint8Slot(4, invoke, 0)


def AddInvoke(builder, invoke):
    RegisterAddInvoke(builder, invoke)


def RegisterAddConcurrency(builder, concurrency):
    builder.PrependUint16Slot(5, concurrency, 0)


def AddConcurrency(builder, concurrency):
    RegisterAddConcurrency(builder, concurrency)


def RegisterAddForceReregister(builder, forceReregister):
    builder.PrependBoolSlot(6, forceReregister, 0)


def AddForceReregister(builder, forceReregister):
    RegisterAddForceReregister(builder, forceReregister)


def RegisterEnd(builder):
    return builder.EndObject()


def End(builder):
    return RegisterEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import asyncio
from asyncio import Future, iscoroutine
from collections import deque
from typing import Optional

import txaio

txaio.use_asyncio()  # noqa

from autobahn.asyncio.util import create_transport_details, transport_channel_id
from autobahn.util import hltype, public
from autobahn.wamp import websocket
from autobahn.websocket import protocol

__all__ = (
    "WampWebSocketClientFactory",
    "WampWebSocketClientProtocol",
    "WampWebSocketServerFactory",
    "WampWebSocketServerProtocol",
    "WebSocketClientFactory",
    "WebSocketClientProtocol",
    "WebSocketServerFactory",
    "WebSocketServerProtocol",
)


def yields(value):
    """
    Returns ``True`` iff the value yields.

    .. seealso:: http://stackoverflow.com/questions/20730248/maybedeferred-analog-with-asyncio
    """
    return isinstance(value, Future) or iscoroutine(value)


class WebSocketAdapterProtocol(asyncio.Protocol):
    """
    Adapter class for asyncio-based WebSocket client and server protocols.
    """

    log = txaio.make_logger()

    peer: str | None = None
    is_server: bool | None = None

    def connection_made(self, transport):
        # asyncio networking framework entry point, called by asyncio
        # when the connection is established (either a client or a server)
        self.log.debug(
            "{func}(transport={transport})",
            func=hltype(self.connection_made),
            transport=transport,
        )

        self.transport = transport

        # determine preliminary transport details (what is know at this point)
        self._transport_details = create_transport_details(
            self.transport, self.is_server
        )

        # backward compatibility
        self.peer = self._transport_details.peer

        self.receive_queue = deque()
        self._consume()

        self._connectionMade()

    def connection_lost(self, exc):
        self._connectionLost(exc)
        # according to asyncio docs, connection_lost(None) is called
        # if something else called transport.close()
        if exc is not None:
            self.transport.close()
        self.transport = None

    def _consume(self):
        self.waiter = Future(loop=self.factory.loop or txaio.config.loop)

        def process(_):
            while self.receive_queue:
                data = self.receive_queue.popleft()
                if self.transport:
                    self._dataReceived(data)
            self._consume()

        self.waiter.add_done_callback(process)

    def data_received(self, data):
        self.receive_queue.append(data)
        if not self.waiter.done():
            self.waiter.set_result(None)

    def _closeConnection(self, abort=False):
        if abort and hasattr(self.transport, "abort"):
            self.transport.abort()
        else:
            self.transport.close()

    def _onOpen(self):
        if self._transport_details.is_secure:
            # now that the TLS opening handshake is complete, the actual TLS channel ID
            # will be available. make sure to set it!
            channel_id = {
                "tls-unique": transport_channel_id(
                    self.transport, self._transport_details.is_server, "tls-unique"
                ),
            }
            self._transport_details.channel_id = channel_id

        res = self.onOpen()
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageBegin(self, isBinary):
        res = self.onMessageBegin(isBinary)
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageFrameBegin(self, length):
        res = self.onMessageFrameBegin(length)
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageFrameData(self, payload):
        res = self.onMessageFrameData(payload)
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageFrameEnd(self):
        res = self.onMessageFrameEnd()
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageFrame(self, payload):
        res = self.onMessageFrame(payload)
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessageEnd(self):
        res = self.onMessageEnd()
        if yields(res):
            asyncio.ensure_future(res)

    def _onMessage(self, payload, isBinary):
        res = self.onMessage(payload, isBinary)
        if yields(res):
            asyncio.ensure_future(res)

    def _onPing(self, payload):
        res = self.onPing(payload)
        if yields(res):
            asyncio.ensure_future(res)

    def _onPong(self, payload):
        res = self.onPong(payload)
        if yields(res):
            asyncio.ensure_future(res)

    def _onClose(self, wasClean, code, reason):
        res = self.onClose(wasClean, code, reason)
        if yields(res):
            asyncio.ensure_future(res)

    def registerProducer(self, producer, streaming):
        raise Exception("not implemented")

    def unregisterProducer(self):
        # note that generic websocket/protocol.py code calls
        # .unregisterProducer whenever we dropConnection -- that's
        # correct behavior on Twisted so either we'd have to
        # try/except there, or special-case Twisted, ..or just make
        # this "not an error"
        pass


@public
class WebSocketServerProtocol(
    WebSocketAdapterProtocol, protocol.WebSocketServerProtocol
):
    """
    Base class for asyncio-based WebSocket server protocols.

    Implements:

    * :class:`autobahn.websocket.interfaces.IWebSocketChannel`
    """

    log = txaio.make_logger()


@public
class WebSocketClientProtocol(
    WebSocketAdapterProtocol, protocol.WebSocketClientProtocol
):
    """
    Base class for asyncio-based WebSocket client protocols.

    Implements:

    * :class:`autobahn.websocket.interfaces.IWebSocketChannel`
    """

    log = txaio.make_logger()

    def _onConnect(self, response):
        res = self.onConnect(response)
        self.log.debug("{func}: {res}", func=hltype(self._onConnect), res=res)
        if yields(res):
            asyncio.ensure_future(res)

    def startTLS(self):
        raise Exception("WSS over explicit proxies not implemented")


class WebSocketAdapterFactory:
    """
    Adapter class for asyncio-based WebSocket client and server factories.
    """

    log = txaio.make_logger()

    def __call__(self):
        proto = self.protocol()
        proto.factory = self
        return proto


@public
class WebSocketServerFactory(WebSocketAdapterFactory, protocol.WebSocketServerFactory):
    """
    Base class for asyncio-based WebSocket server factories.

    Implements:

    * :class:`autobahn.websocket.interfaces.IWebSocketServerChannelFactory`
    """

    log = txaio.make_logger()

    protocol = WebSocketServerProtocol

    def __init__(self, *args, **kwargs):
        """
        .. note::
            In addition to all arguments to the constructor of
            :meth:`autobahn.websocket.interfaces.IWebSocketServerChannelFactory`,
            you can supply a ``loop`` keyword argument to specify the
            asyncio event loop to be used.
        """
        loop = kwargs.pop("loop", None)
        self.loop = loop or asyncio.get_event_loop()

        protocol.WebSocketServerFactory.__init__(self, *args, **kwargs)


@public
class WebSocketClientFactory(WebSocketAdapterFactory, protocol.WebSocketClientFactory):
    """
    Base class for asyncio-based WebSocket client factories.

    Implements:

    * :class:`autobahn.websocket.interfaces.IWebSocketClientChannelFactory`
    """

    log = txaio.make_logger()

    def __init__(self, *args, **kwargs):
        """

        .. note::
            In addition to all arguments to the constructor of
            :meth:`autobahn.websocket.interfaces.IWebSocketClientChannelFactory`,
            you can supply a ``loop`` keyword argument to specify the
            asyncio event loop to be used.
        """
        loop = kwargs.pop("loop", None)
        self.loop = loop or asyncio.get_event_loop()

        protocol.WebSocketClientFactory.__init__(self, *args, **kwargs)


@public
class WampWebSocketServerProtocol(
    websocket.WampWebSocketServerProtocol, WebSocketServerProtocol
):
    """
    asyncio-based WAMP-over-WebSocket server protocol.

    Implements:

    * :class:`autobahn.wamp.interfaces.ITransport`
    """

    log = txaio.make_logger()


@public
class WampWebSocketServerFactory(
    websocket.WampWebSocketServerFactory, WebSocketServerFactory
):
    """
    asyncio-based WAMP-over-WebSocket server factory.
    """

    log = txaio.make_logger()

    protocol = WampWebSocketServerProtocol

    def __init__(self, factory, *args, **kwargs):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializers: A list of WAMP serializers to use (or ``None``
            for all available serializers).
        :type serializers: list of objects implementing
            :class:`autobahn.wamp.interfaces.ISerializer`
        """

        serializers = kwargs.pop("serializers", None)

        websocket.WampWebSocketServerFactory.__init__(self, factory, serializers)

        kwargs["protocols"] = self._protocols

        # noinspection PyCallByClass
        WebSocketServerFactory.__init__(self, *args, **kwargs)


@public
class WampWebSocketClientProtocol(
    websocket.WampWebSocketClientProtocol, WebSocketClientProtocol
):
    """
    asyncio-based WAMP-over-WebSocket client protocols.

    Implements:

    * :class:`autobahn.wamp.interfaces.ITransport`
    """

    log = txaio.make_logger()


@public
class WampWebSocketClientFactory(
    websocket.WampWebSocketClientFactory, WebSocketClientFactory
):
    """
    asyncio-based WAMP-over-WebSocket client factory.
    """

    log = txaio.make_logger()

    protocol = WampWebSocketClientProtocol

    def __init__(self, factory, *args, **kwargs):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializer: The WAMP serializer to use (or ``None`` for
           "best" serializer, chosen as the first serializer available from
           this list: CBOR, MessagePack, UBJSON, JSON).
        :type serializer: object implementing :class:`autobahn.wamp.interfaces.ISerializer`
        """

        serializers = kwargs.pop("serializers", None)

        websocket.WampWebSocketClientFactory.__init__(self, factory, serializers)

        kwargs["protocols"] = self._protocols

        WebSocketClientFactory.__init__(self, *args, **kwargs)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

import autobahn
from autobahn import flatbuffers

# zlmdb is not a hard dependency of autobahn; the cross-project version-sync
# check (and its test) only run when zlmdb is also installed.
try:
    import zlmdb.flatbuffers  # noqa: F401

    HAS_ZLMDB = True
except ImportError:
    HAS_ZLMDB = False


class TestFlatBuffersVersion(unittest.TestCase):
    def test_version_tuple(self):
        "autobahn.flatbuffers.version() returns a parsed 5-tuple"
        v = flatbuffers.version()
        self.assertIsInstance(v, tuple)
        self.assertEqual(len(v), 5)
        major, minor, patch, commits, commit_hash = v
        self.assertIsInstance(major, int)
        self.assertIsInstance(minor, int)
        self.assertIsInstance(patch, int)
        # a released, tagged runtime has no extra commits/hash
        self.assertTrue(commits is None or isinstance(commits, int))
        self.assertTrue(commit_hash is None or isinstance(commit_hash, str))

    def test_version_falls_back_to_dunder_version_on_bare_hash(self):
        "version() falls back to __version__ when __git_version__ is a non-matching bare hash"
        # On built wheels / shallow clones `git describe --tags --always` yields a bare
        # commit hash that does NOT match the vX.Y.Z pattern; version() must fall back to
        # the reliable static __version__, NOT return (0, 0, 0, None, None).
        orig = flatbuffers.__git_version__
        try:
            flatbuffers.__git_version__ = "19b2300f"
            v = flatbuffers.version()
            self.assertNotEqual(v, (0, 0, 0, None, None))
            expected = tuple(int(x) for x in flatbuffers.__version__.split(".")[:3])
            self.assertEqual(v, (*expected, None, None))
        finally:
            flatbuffers.__git_version__ = orig

    def test_version_falls_back_to_dunder_version_on_unknown(self):
        "version() falls back to __version__ when __git_version__ is 'unknown'"
        orig = flatbuffers.__git_version__
        try:
            flatbuffers.__git_version__ = "unknown"  # hatch_build default on describe failure
            expected = tuple(int(x) for x in flatbuffers.__version__.split(".")[:3])
            self.assertEqual(flatbuffers.version(), (*expected, None, None))
        finally:
            flatbuffers.__git_version__ = orig

    def test_version_preserves_git_describe_detail(self):
        "version() still returns rich git-describe detail when __git_version__ matches"
        orig = flatbuffers.__git_version__
        try:
            flatbuffers.__git_version__ = "v25.9.23-71-g19b2300f"
            self.assertEqual(flatbuffers.version(), (25, 9, 23, 71, "19b2300f"))
        finally:
            flatbuffers.__git_version__ = orig

    @unittest.skipUnless(HAS_ZLMDB, "zlmdb not installed")
    def test_zlmdb_flatbuffers_in_sync(self):
        "autobahn and zlmdb vendor the same FlatBuffers version (data-in-transit vs data-at-rest)"
        version = autobahn.check_zlmdb_flatbuffers_version_in_sync()
        self.assertEqual(version, flatbuffers.__version__)
        import zlmdb.flatbuffers

        self.assertEqual(version, zlmdb.flatbuffers.__version__)

    @unittest.skipUnless(HAS_ZLMDB, "zlmdb not installed")
    def test_zlmdb_flatbuffers_mismatch_raises(self):
        "check_zlmdb_flatbuffers_version_in_sync() raises when the two vendored __version__ differ"
        # guard against regressions to comparing version() (which is (0,0,0,...) on
        # built wheels and would never detect a mismatch): force the two vendored
        # __version__ strings apart and assert the check actually raises.
        orig = flatbuffers.__version__
        try:
            flatbuffers.__version__ = orig + "-mismatch"
            with self.assertRaises(RuntimeError):
                autobahn.check_zlmdb_flatbuffers_version_in_sync()
        finally:
            flatbuffers.__version__ = orig



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# t.i.reactor doesn't exist until we've imported it once, but we
# need it to exist so we can @patch it out in the tests ...
from twisted.internet import reactor  # noqa
from unittest.mock import patch, Mock

from twisted.internet.defer import inlineCallbacks, succeed
from twisted.trial import unittest

from autobahn.twisted.wamp import ApplicationRunner


def raise_error(*args, **kw):
    raise RuntimeError("we always fail")


class TestApplicationRunner(unittest.TestCase):
    @patch("twisted.internet.reactor")
    def test_runner_default(self, fakereactor):
        fakereactor.connectTCP = Mock(side_effect=raise_error)
        runner = ApplicationRunner("ws://fake:1234/ws", "dummy realm")

        # we should get "our" RuntimeError when we call run
        self.assertRaises(RuntimeError, runner.run, raise_error)

        # both reactor.run and reactor.stop should have been called
        self.assertEqual(fakereactor.run.call_count, 1)
        self.assertEqual(fakereactor.stop.call_count, 1)

    @patch("twisted.internet.reactor")
    @inlineCallbacks
    def test_runner_no_run(self, fakereactor):
        fakereactor.connectTCP = Mock(side_effect=raise_error)
        runner = ApplicationRunner("ws://fake:1234/ws", "dummy realm")

        try:
            yield runner.run(raise_error, start_reactor=False)
            self.fail()  # should have raise an exception, via Deferred

        except RuntimeError as e:
            # make sure it's "our" exception
            self.assertEqual(e.args[0], "we always fail")

        # neither reactor.run() NOR reactor.stop() should have been called
        # (just connectTCP() will have been called)
        self.assertEqual(fakereactor.run.call_count, 0)
        self.assertEqual(fakereactor.stop.call_count, 0)

    @patch("twisted.internet.reactor")
    def test_runner_no_run_happypath(self, fakereactor):
        proto = Mock()
        fakereactor.connectTCP = Mock(return_value=succeed(proto))
        runner = ApplicationRunner("ws://fake:1234/ws", "dummy realm")

        d = runner.run(Mock(), start_reactor=False)

        # shouldn't have actually connected to anything
        # successfully, and the run() call shouldn't have inserted
        # any of its own call/errbacks. (except the cleanup handler)
        self.assertFalse(d.called)
        self.assertEqual(1, len(d.callbacks))

        # neither reactor.run() NOR reactor.stop() should have been called
        # (just connectTCP() will have been called)
        self.assertEqual(fakereactor.run.call_count, 0)
        self.assertEqual(fakereactor.stop.call_count, 0)

    @patch("twisted.internet.reactor")
    def test_runner_bad_proxy(self, fakereactor):
        proxy = "myproxy"

        self.assertRaises(
            AssertionError,
            ApplicationRunner,
            "ws://fake:1234/ws",
            "dummy realm",
            proxy=proxy,
        )

    @patch("twisted.internet.reactor")
    def test_runner_proxy(self, fakereactor):
        proto = Mock()
        fakereactor.connectTCP = Mock(return_value=succeed(proto))

        proxy = {"host": "myproxy", "port": 3128}

        runner = ApplicationRunner("ws://fake:1234/ws", "dummy realm", proxy=proxy)

        d = runner.run(Mock(), start_reactor=False)

        # shouldn't have actually connected to anything
        # successfully, and the run() call shouldn't have inserted
        # any of its own call/errbacks. (except the cleanup handler)
        self.assertFalse(d.called)
        self.assertEqual(1, len(d.callbacks))

        # neither reactor.run() NOR reactor.stop() should have been called
        # (just connectTCP() will have been called)
        self.assertEqual(fakereactor.run.call_count, 0)
        self.assertEqual(fakereactor.stop.call_count, 0)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class RouterRoles(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RouterRoles()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRouterRoles(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # RouterRoles
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RouterRoles
    def Broker(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.BrokerFeatures import BrokerFeatures

            obj = BrokerFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RouterRoles
    def Dealer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.DealerFeatures import DealerFeatures

            obj = DealerFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


def RouterRolesStart(builder):
    builder.StartObject(2)


def Start(builder):
    RouterRolesStart(builder)


def RouterRolesAddBroker(builder, broker):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(broker), 0
    )


def AddBroker(builder, broker):
    RouterRolesAddBroker(builder, broker)


def RouterRolesAddDealer(builder, dealer):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(dealer), 0
    )


def AddDealer(builder, dealer):
    RouterRolesAddDealer(builder, dealer)


def RouterRolesEnd(builder):
    return builder.EndObject()


def End(builder):
    return RouterRolesEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto


class AnyMessage(object):
    NONE = 0
    Hello = 1
    Welcome = 2
    Abort = 3
    Challenge = 4
    Authenticate = 5
    Goodbye = 6
    Error = 7
    Publish = 8
    Published = 9
    Subscribe = 10
    Subscribed = 11
    Unsubscribe = 12
    Unsubscribed = 13
    Event = 14
    EventReceived = 15
    Call = 16
    Cancel = 17
    Result = 18
    Register = 19
    Registered = 20
    Unregister = 21
    Unregistered = 22
    Invocation = 23
    Interrupt = 24
    Yield = 25



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest
from unittest.mock import Mock

from autobahn.testutil import FakeTransport
from autobahn.twisted.rawsocket import (
    WampRawSocketClientFactory,
    WampRawSocketClientProtocol,
    WampRawSocketServerFactory,
    WampRawSocketServerProtocol,
)


class RawSocketHandshakeTests(unittest.TestCase):
    def test_handshake_succeeds(self):
        """
        A client can connect to a server.
        """
        session_mock = Mock()
        t = FakeTransport()
        f = WampRawSocketClientFactory(lambda: session_mock)
        p = WampRawSocketClientProtocol()
        p.transport = t
        p.factory = f

        server_session_mock = Mock()
        st = FakeTransport()
        sf = WampRawSocketServerFactory(lambda: server_session_mock)
        sp = WampRawSocketServerProtocol()
        sp.transport = st
        sp.factory = sf

        sp.connectionMade()
        p.connectionMade()

        # Send the server the client handshake
        sp.dataReceived(t._written[0:1])
        sp.dataReceived(t._written[1:4])

        # Send the client the server handshake
        p.dataReceived(st._written)

        # The handshake succeeds, a session on each end is created
        # onOpen is called on the session
        session_mock.onOpen.assert_called_once_with(p)
        server_session_mock.onOpen.assert_called_once_with(sp)

    def test_receive_limit_advertised_and_enforced(self):
        """
        ``setProtocolOptions(maxMessagePayloadSize=...)`` configures the server's
        received-message size cap: the advertised handshake length exponent and
        the enforced ``MAX_LENGTH`` are the configured max rounded up to the next
        power of two. This mirrors the asyncio backend (see
        ``test_aio_rawsocket.RECV_LIMIT_CASES``) so both backends make identical
        accept/reject decisions for the same configuration (#1911).
        """
        # (maxMessagePayloadSize, advertised length exponent, enforced cap)
        cases = [
            (512, 0, 512),
            (1000, 1, 1024),  # rounded up to the next power of two
            (1024, 1, 1024),
            (4096, 3, 4096),
            (2**20, 11, 2**20),
            (2**24, 15, 2**24),
        ]
        for max_size, exp, cap in cases:
            with self.subTest(max_size=max_size):
                sf = WampRawSocketServerFactory(lambda: Mock())
                sf.setProtocolOptions(maxMessagePayloadSize=max_size)
                sp = sf.buildProtocol(None)
                sp.transport = FakeTransport()
                sp.connectionMade()
                ser_id = sorted(sf._serializers.keys())[0]
                # client opening handshake: magic, (length-exp 15 | serializer)
                sp.dataReceived(bytes([0x7F, 0xF0 | ser_id, 0, 0]))
                written = sp.transport._written
                self.assertEqual(written[1] >> 4, exp)
                self.assertEqual(sp.MAX_LENGTH, cap)

    def test_server_bad_magic_byte_aborts_cleanly(self):
        """
        A server receiving an invalid magic byte in the opening handshake
        aborts the transport cleanly instead of raising ``TransportLost`` out
        of ``dataReceived`` (regression test for #1850).
        """
        server_session_mock = Mock()
        st = FakeTransport()
        sf = WampRawSocketServerFactory(lambda: server_session_mock)
        sp = WampRawSocketServerProtocol()
        sp.transport = st
        sp.factory = sf

        sp.connectionMade()

        # any non-0x7f first octet is an invalid magic byte; this must not raise
        sp.dataReceived(b"GET ")

        # the transport was aborted and no WAMP session was started
        self.assertTrue(st.abort_called())
        server_session_mock.onOpen.assert_not_called()

    def test_client_bad_magic_byte_aborts_cleanly(self):
        """
        A client receiving an invalid magic byte in the opening handshake
        aborts the transport cleanly instead of raising ``TransportLost`` out
        of ``dataReceived`` (regression test for #1850).
        """
        session_mock = Mock()
        t = FakeTransport()
        f = WampRawSocketClientFactory(lambda: session_mock)
        p = WampRawSocketClientProtocol()
        p.transport = t
        p.factory = f

        p.connectionMade()

        # any non-0x7f first octet is an invalid magic byte; this must not raise
        p.dataReceived(b"GET ")

        # the transport was aborted and no WAMP session was started
        self.assertTrue(t.abort_called())
        session_mock.onOpen.assert_not_called()



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Git version from deps/flatbuffers submodule.
# This file is regenerated at build time by hatch_build.py.
# The version is captured via `git describe --tags` in the submodule.
#
# Format: "v25.9.23" (tagged release) or "v25.9.23-2-g95053e6a" (post-tag)
#
# If building from sdist without git, this will retain the version
# from when the sdist was created.

__git_version__ = "v25.12.19"



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# IHostnameResolver et al. were added in Twisted 17.1.0 .. before
# that, it was IResolverSimple only.

try:
    from twisted.internet.interfaces import IHostnameResolver
except ImportError:
    raise ImportError("Twisted 17.1.0 or later required for autobahn.twisted.testing")

from twisted.internet._resolver import (
    HostResolution,
)  # "internal" class, but it's simple
from twisted.internet.address import IPv4Address
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import ISSLTransport

try:
    from twisted.internet.testing import MemoryReactorClock
except ImportError:
    from twisted.test.proto_helpers import MemoryReactorClock
from zope.interface import directlyProvides, implementer

from autobahn.twisted.websocket import (
    WebSocketServerFactory,
    WebSocketServerProtocol,
    _TwistedWebSocketClientAgent,
)
from autobahn.websocket.interfaces import IWebSocketClientAgent
from twisted.test import iosim

__all__ = (
    "MemoryReactorClockResolver",
    "create_memory_agent",
    "create_pumper",
)


@implementer(IHostnameResolver)
class _StaticTestResolver:
    def resolveHostName(self, receiver, hostName, portNumber=0):
        """
        Implement IHostnameResolver which always returns 127.0.0.1:31337
        """
        resolution = HostResolution(hostName)
        receiver.resolutionBegan(resolution)
        receiver.addressResolved(
            IPv4Address("TCP", "127.0.0.1", 31337 if portNumber == 0 else portNumber)
        )
        receiver.resolutionComplete()


# in previous revisions, we exported MemoryReactorClockResolver so
# this maintains compatibility with any downstream code
MemoryReactorClockResolver = MemoryReactorClock


class _TwistedWebMemoryAgent(IWebSocketClientAgent):
    """
    A testing agent which will hook up an instance of
    `server_protocol` for every client that is created via the `open`
    API call.

    :param reactor: the reactor to use for tests (usually an instance
        of MemoryReactorClockResolver)

    :param pumper: an implementation IPumper (e.g. as returned by
        `create_pumper`)

    :param server_protocol: the server-side WebSocket protocol class
        to instantiate (e.g. a subclass of `WebSocketServerProtocol`
    """

    def __init__(self, reactor, pumper, server_protocol):
        self._reactor = reactor
        self._server_protocol = server_protocol
        self._pumper = pumper

        # our "real" underlying agent under test
        self._agent = _TwistedWebSocketClientAgent(self._reactor)
        self._pumps = set()
        self._servers = dict()  # client -> server

    def open(self, transport_config, options, protocol_class=None):
        """
        Implement IWebSocketClientAgent with in-memory transports.

        :param transport_config: a string starting with 'wss://' or
            'ws://'

        :param options: a dict containing options

        :param protocol_class: the client protocol class to
            instantiate (or `None` for defaults, which is to use
            `WebSocketClientProtocol`)
        """
        is_secure = transport_config.startswith("wss://")

        # call our "real" agent
        real_client_protocol = self._agent.open(
            transport_config,
            options,
            protocol_class=protocol_class,
        )

        if is_secure:
            host, port, factory, context_factory, timeout, bindAddress = (
                self._reactor.sslClients[-1]
            )
        else:
            host, port, factory, timeout, bindAddress = self._reactor.tcpClients[-1]

        server_address = IPv4Address("TCP", "127.0.0.1", port)
        client_address = IPv4Address("TCP", "127.0.0.1", 31337)

        server_protocol = self._server_protocol()
        # the protocol could already have a factory
        if getattr(server_protocol, "factory", None) is None:
            server_protocol.factory = WebSocketServerFactory()

        server_transport = iosim.FakeTransport(
            server_protocol,
            isServer=True,
            hostAddress=server_address,
            peerAddress=client_address,
        )
        clientProtocol = factory.buildProtocol(None)
        client_transport = iosim.FakeTransport(
            clientProtocol,
            isServer=False,
            hostAddress=client_address,
            peerAddress=server_address,
        )

        if is_secure:
            directlyProvides(server_transport, ISSLTransport)
            directlyProvides(client_transport, ISSLTransport)

        pump = iosim.connect(
            server_protocol, server_transport, clientProtocol, client_transport
        )
        self._pumper.add(pump)

        def add_mapping(proto):
            self._servers[proto] = server_protocol
            return proto

        real_client_protocol.addCallback(add_mapping)
        return real_client_protocol


class _Kalamazoo:
    """
    Feeling whimsical about class names, see https://en.wikipedia.org/wiki/Handcar

    This is 'an IOPump pumper', an object which causes a series of
    IOPumps it is monitoring to do their I/O operations
    periodically. This needs the 'real' reactor which trial drives,
    because reasons:

     - so @inlineCallbacks / async-def functions work
       (if I could explain exactly why here, I would)

     - we need to 'break the loop' of synchronous calls somewhere and
       polluting the tests themselves with that is bad

     - get rid of e.g. .flush() calls in tests themselves (thus
      'teaching' the tests about details of I/O scheduling that they
      shouldn't know).
    """

    def __init__(self):
        self._pumps = set()
        self._pumping = False
        self._waiting_for_stop = []
        from twisted.internet import reactor as global_reactor

        self._global_reactor = global_reactor

    def add(self, p):
        """
        Add a new IOPump. It will be removed when both its client and
        server are disconnected.
        """
        self._pumps.add(p)

    def start(self):
        """
        Begin triggering I/O in all IOPump instances we have. We will keep
        periodically 'pumping' our IOPumps until `.stop()` is
        called. Call from `setUp()` for example.
        """
        if self._pumping:
            return
        self._pumping = True
        self._global_reactor.callLater(0, self._pump_once)

    def stop(self):
        """
        :returns: a Deferred that fires when we have stopped pump()-ing

        Call from `tearDown()`, for example.
        """
        if self._pumping or len(self._waiting_for_stop):
            d = Deferred()
            self._waiting_for_stop.append(d)
            self._pumping = False
            return d
        d = Deferred()
        d.callback(None)
        return d

    def _pump_once(self):
        """
        flush all data from all our IOPump instances and schedule another
        iteration on the global reactor
        """
        if self._pumping:
            self._flush()
            self._global_reactor.callLater(0.1, self._pump_once)
        else:
            for d in self._waiting_for_stop:
                d.callback(None)
            self._waiting_for_stop = []

    def _flush(self):
        """
        Flush all data between pending client/server pairs.
        """
        old_pumps = self._pumps
        new_pumps = self._pumps = set()
        for p in old_pumps:
            p.flush()
            if p.clientIO.disconnected and p.serverIO.disconnected:
                continue
            new_pumps.add(p)


def create_pumper():
    """
    return a new instance implementing IPumper
    """
    return _Kalamazoo()


def create_memory_agent(reactor, pumper, server_protocol):
    """
    return a new instance implementing `IWebSocketClientAgent`.

    connection attempts will be satisfied by traversing the Upgrade
    request path starting at `resource` to find a `WebSocketResource`
    and then exchange data between client and server using purely
    in-memory buffers.
    """
    # Note, we currently don't actually do any "resource traversing"
    # and basically accept any path at all to our websocket resource
    if server_protocol is None:
        server_protocol = WebSocketServerProtocol
    return _TwistedWebMemoryAgent(reactor, pumper, server_protocol)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class Error(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Error()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsError(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Error
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Error
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Error
    def RequestType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # Error
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Error
    def Error(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Positional values for application-defined error details.
    # Error
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Error
    def ArgsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Error
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Error
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Keyword values for application-defined error details.
    # Error
    def Kwargs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Error
    def KwargsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Error
    def KwargsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Error
    def KwargsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.
    # Error
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Error
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Error
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Error
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # The specific scheme in use with Payload Passthru (PPT) mode for the application payload.
    # Error
    def PptScheme(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The specific serializer encoding the application payload with the Payload Passthru (PPT) scheme in use.
    # Error
    def PptSerializer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The cryptographic algorithm ("cipher") encrypting the application payload with the Payload Passthru (PPT) scheme in use.
    # Error
    def PptCipher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The identifier or reference to the encryption key that was used to encrypt the payload with the Payload Passthru (PPT) scheme and cipher in use.
    # Error
    def PptKeyid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Error
    def ForwardFor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from autobahn.wamp.gen.wamp.proto.Principal import Principal

            obj = Principal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Error
    def ForwardForLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Error
    def ForwardForIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0


def ErrorStart(builder):
    builder.StartObject(12)


def Start(builder):
    ErrorStart(builder)


def ErrorAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    ErrorAddSession(builder, session)


def ErrorAddRequestType(builder, requestType):
    builder.PrependUint16Slot(1, requestType, 0)


def AddRequestType(builder, requestType):
    ErrorAddRequestType(builder, requestType)


def ErrorAddRequest(builder, request):
    builder.PrependUint64Slot(2, request, 0)


def AddRequest(builder, request):
    ErrorAddRequest(builder, request)


def ErrorAddError(builder, error):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(error), 0
    )


def AddError(builder, error):
    ErrorAddError(builder, error)


def ErrorAddArgs(builder, args):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0
    )


def AddArgs(builder, args):
    ErrorAddArgs(builder, args)


def ErrorStartArgsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArgsVector(builder, numElems):
    return ErrorStartArgsVector(builder, numElems)


def ErrorAddKwargs(builder, kwargs):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(kwargs), 0
    )


def AddKwargs(builder, kwargs):
    ErrorAddKwargs(builder, kwargs)


def ErrorStartKwargsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKwargsVector(builder, numElems):
    return ErrorStartKwargsVector(builder, numElems)


def ErrorAddPayload(builder, payload):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0
    )


def AddPayload(builder, payload):
    ErrorAddPayload(builder, payload)


def ErrorStartPayloadVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayloadVector(builder, numElems):
    return ErrorStartPayloadVector(builder, numElems)


def ErrorAddPptScheme(builder, pptScheme):
    builder.PrependUint8Slot(7, pptScheme, 0)


def AddPptScheme(builder, pptScheme):
    ErrorAddPptScheme(builder, pptScheme)


def ErrorAddPptSerializer(builder, pptSerializer):
    builder.PrependUint8Slot(8, pptSerializer, 0)


def AddPptSerializer(builder, pptSerializer):
    ErrorAddPptSerializer(builder, pptSerializer)


def ErrorAddPptCipher(builder, pptCipher):
    builder.PrependUint8Slot(9, pptCipher, 0)


def AddPptCipher(builder, pptCipher):
    ErrorAddPptCipher(builder, pptCipher)


def ErrorAddPptKeyid(builder, pptKeyid):
    builder.PrependUOffsetTRelativeSlot(
        10, flatbuffers.number_types.UOffsetTFlags.py_type(pptKeyid), 0
    )


def AddPptKeyid(builder, pptKeyid):
    ErrorAddPptKeyid(builder, pptKeyid)


def ErrorAddForwardFor(builder, forwardFor):
    builder.PrependUOffsetTRelativeSlot(
        11, flatbuffers.number_types.UOffsetTFlags.py_type(forwardFor), 0
    )


def AddForwardFor(builder, forwardFor):
    ErrorAddForwardFor(builder, forwardFor)


def ErrorStartForwardForVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartForwardForVector(builder, numElems):
    return ErrorStartForwardForVector(builder, numElems)


def ErrorEnd(builder):
    return builder.EndObject()


def End(builder):
    return ErrorEnd(builder)
